# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from pkg_resources import resource_filename # @UnresolvedImport

import gi
gi.require_version('Gtk', '3.0')  # @UndefinedVariable
from gi.repository import Gtk  # @UnresolvedImport

from matplotlib.figure import Figure
from matplotlib.backends.backend_gtk3cairo import FigureCanvasGTK3Cairo as FigureCanvasGTK

from pyxrd.generic.views import ObjectListStoreView, DialogView, BaseView

class EditMarkerView(BaseView):
    builder = resource_filename(__name__, "../glade/edit_marker.glade")
    top = "edit_marker"

    widget_format = "marker_%s"

    def __init__(self, *args, **kwargs):
        BaseView.__init__(self, *args, **kwargs)

        self.parent.set_title("Edit Markers")

class EditMarkersView(ObjectListStoreView):
    extra_widget_builder = resource_filename(__name__, "../glade/find_peaks.glade")
    extra_widget_toplevel = "vbox_find_peaks"
    resizable = False

    def __init__(self, *args, **kwargs):
        ObjectListStoreView.__init__(self, *args, **kwargs)

        self[self.subview_toplevel].child_set_property(self["frame_object_param"], "resize", False)

    def set_selection_state(self, value):
        super(EditMarkersView, self).set_selection_state(value)
        self["cmd_match_minerals"].set_sensitive(value is not None)

    pass # end of class

class MatchMineralsView(DialogView):
    title = "Match minerals"
    subview_builder = resource_filename(__name__, "../glade/match_minerals.glade")
    subview_toplevel = "tbl_match_minerals"
    modal = True

    def __init__(self, *args, **kwargs):
        DialogView.__init__(self, *args, **kwargs)
        self.tv_minerals = self["tv_minerals"]
        self.tv_matches = self["tv_matches"]

    pass # end of class

class DetectPeaksView(DialogView):
    title = "Auto detect peaks"
    subview_builder = resource_filename(__name__, "../glade/find_peaks_dialog.glade")
    subview_toplevel = "tbl_find_peaks"
    modal = True
    resizable = False

    widget_groups = {
        'full_mode_only': [
            "pattern",
            "lbl_pattern",
            "hseparator1"
        ]
    }

    def __init__(self, *args, **kwargs):
        DialogView.__init__(self, *args, **kwargs)

        self.graph_parent = self["view_graph"]
        self.setup_matplotlib_widget()

    def setup_matplotlib_widget(self):
        #style = Gtk.Style()
        self.figure = Figure(dpi=72) #, edgecolor=str(style.bg[2]), facecolor=str(style.bg[2]))

        self.plot = self.figure.add_subplot(111)
        self.plot.set_ylabel('# of peaks', labelpad=1)
        self.plot.set_xlabel('Threshold', labelpad=1)
        self.plot.autoscale_view()

        self.figure.subplots_adjust(left=0.15, right=0.875, top=0.875, bottom=0.15)
        self.matlib_canvas = FigureCanvasGTK(self.figure)

        self.graph_parent.add(self.matlib_canvas)
        self.graph_parent.show_all()

    pass # end of class


# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from .specimens import SpecimenView
from .markers import (
    DetectPeaksView,
    EditMarkersView,
    EditMarkerView,
    MatchMineralsView
)

__all__ = [
    "SpecimenView",
    "DetectPeaksView",
    "EditMarkersView",
    "EditMarkerView",
    "MatchMineralsView"
]


#!/usr/bin/python

# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import unittest

from pyxrd.generic.models import PyXRDModel

__all__ = [
    'TestPyXRDModel',
]

class TestPyXRDModel(unittest.TestCase):

    def setUp(self):
        self.pyxrdmodel = PyXRDModel()

    def tearDown(self):
        del self.pyxrdmodel

    def test_not_none(self):
        self.assertIsNotNone(self.pyxrdmodel)

    # TODO everything else

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import logging
logger = logging.getLogger(__name__)

from .refine_method_meta import RefineMethodMeta
from .refine_async_helper import RefineAsyncHelper

from pyxrd.calculations.mixture import get_optimized_residual

class RefineMethod(RefineAsyncHelper, metaclass=RefineMethodMeta):
    
    """
        The `RefineMethod` class is the base class for refinement methods.
        Sub-classes will be registered in the metaclass.
    """

    name = "Name of the algorithm"

    description = "A slightly longer explanation of algorithm"

    # The value of this index is important;
    # Some ranges are reserved to prevent immediate overlaps:
    #  - negative values should not be used (not enforced)
    #  - the range 0 - 999 is reserved for built-in methods
    #  - all other values can be used for third-party methods, it is up to the
    #    final user to check if they don't overlap. If these methods become
    #    a built-in method, they'll receive a new index in the preserved range
    index = -1

    disabled = True
    
    residual_callback = property(fget=lambda *s: get_optimized_residual)

    def __call__(self, refiner, stop=None, **kwargs):

        self._stop = stop

        options = self.get_options()
        for arg in self.options:
            options[arg] = kwargs.get(arg, getattr(self, arg))

        return self.run(refiner, **options)

    def run(self, refiner, **kwargs):
        raise NotImplementedError("The run method of RefineRun should be implemented by sub-classes...")

    def get_options(self):
        """ Returns a dict containing the option attribute names as keys and
        their values as values """
        return { name: getattr(self, name) for name in self.options }

    pass #end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from itertools import product, combinations

import numpy as np

from ..refine_method import RefineMethod
from ..refine_method_option import RefineMethodOption

class RefineBruteForceRun(RefineMethod):
    name = "Brute force algorithm"
    description = "Refinement using a Brute Force algorithm"
    index = 3
    disabled = False

    num_samples = RefineMethodOption('Number of samples', 11, [3, 1000], int)

    def run(self, refiner, num_samples=11, stop=None, **kwargs):
        """
            Refinement using a Brute Force algorithm
        """

        self.refiner = refiner

        num_params = len(refiner.ranges)

        npbounds = np.array(refiner.ranges, dtype=float)
        npmins = npbounds[:, 0]
        npranges = npbounds[:, 1] - npbounds[:, 0]

        def generate():
            # Generates the solutions for async evaluation
            if num_params == 1:
                for index in range(num_samples):
                    npindex = np.array([index / float(num_samples - 1)], dtype=float)
                    solution = npmins + npranges * npindex
                    yield solution
            else:
                # Generate a grid for each possible combination of parameters:
                for par1, par2 in combinations(list(range(num_params)), 2):
                    # produce the grid indices for those parameters
                    # keep the others half-way their range:
                    indeces = np.ones(shape=(num_params,), dtype=float) * 0.5
                    for par_indeces in product(list(range(num_samples)), repeat=2):
                        indeces[par1] = par_indeces[0] / float(num_samples - 1)
                        indeces[par2] = par_indeces[1] / float(num_samples - 1)
                        # Make the solution:
                        solution = npmins + npranges * indeces
                        yield solution

        self.do_async_evaluation(generate)
        
        refiner.apply_initial_solution()

    pass #end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import zipfile
from random import choice

from mvc.models.properties import (
    StringProperty, SignalMixin, IntegerProperty, ReadOnlyMixin
)

from pyxrd.generic.io import storables, Storable, COMPRESSION
from pyxrd.generic.models import DataModel

from pyxrd.calculations.data_objects import PhaseData
from pyxrd.calculations.phases import get_diffracted_intensity
from pyxrd.file_parsers.json_parser import JSONParser

@storables.register()
class AbstractPhase(DataModel, Storable):

    # MODEL INTEL:
    class Meta(DataModel.Meta):
        store_id = "AbstractPhase"

    _data_object = None
    @property
    def data_object(self):
        return self._data_object

    project = property(DataModel.parent.fget, DataModel.parent.fset)

    # PROPERTIES:

    #: The name of this Phase
    name = StringProperty(
        default="New Phase", text="Name",
        visible=True, persistent=True, tabular=True,
    )

    #: The # of components
    @IntegerProperty(
        default=0, text="# of components",
        visible=True, persistent=True, tabular=True, widget_type="entry",
        mix_with=(ReadOnlyMixin,)
    )
    def G(self):
        return 0

    #: The Reichweite
    @IntegerProperty(
        default=0, text="Reichweite",
        visible=True, persistent=False, tabular=True, widget_type="entry",
        mix_with=(ReadOnlyMixin,)
    )
    def R(self):
        return 0

    #: The color this phase's X-ray diffraction pattern should have.
    display_color = StringProperty(
        default="#FFB600", text="Display color",
        visible=True, persistent=True, tabular=True, widget_type='color',
        signal_name="visuals_changed",
        mix_with=(SignalMixin,)
    )

    line_colors = [
        "#004488",
        "#FF4400",
        "#559911",
        "#770022",
        "#AACC00",
        "#441177",
    ]

    # ------------------------------------------------------------
    #      Initialization and other internals
    # ------------------------------------------------------------
    def __init__(self, *args, **kwargs):

        my_kwargs = self.pop_kwargs(kwargs,
            "data_name", "data_G", "data_R",
            *[prop.label for prop in AbstractPhase.Meta.get_local_persistent_properties()]
        )
        super(AbstractPhase, self).__init__(*args, **kwargs)
        kwargs = my_kwargs

        with self.data_changed.hold():

            self._data_object = PhaseData()

            self.name = self.get_kwarg(kwargs, self.name, "name", "data_name")
            self.display_color = self.get_kwarg(kwargs, choice(self.line_colors), "display_color")

    def __repr__(self):
        return "AbstractPhase(name='%s')" % (self.name)

    # ------------------------------------------------------------
    #      Input/Output stuff
    # ------------------------------------------------------------
    def resolve_json_references(self):
        pass # nothing to do, sub-classes should override

    def _pre_multi_save(self, phases, ordered_phases):
        pass # nothing to do, sub-classes should override

    def _post_multi_save(self):
        pass # nothing to do, sub-classes should override

    @classmethod
    def save_phases(cls, phases, filename):
        """
            Saves multiple phases to a single file.
        """
        ordered_phases = list(phases) # make a copy
        for phase in phases:
            phase._pre_multi_save(phases, ordered_phases)

        with zipfile.ZipFile(filename, 'w', compression=COMPRESSION) as zfile:
            for i, phase in enumerate(ordered_phases):
                zfile.writestr("%d###%s" % (i, phase.uuid), phase.dump_object())

        for phase in ordered_phases:
            phase._post_multi_save()

        # After export we change all the UUID's
        # This way, we're sure that we're not going to import objects with
        # duplicate UUID's!
        type(cls).object_pool.change_all_uuids()

    # ------------------------------------------------------------
    #      Methods & Functions
    # ------------------------------------------------------------
    def get_diffracted_intensity(self, range_theta, range_stl, *args):
        return get_diffracted_intensity(range_theta, range_stl, self.data_object)

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import logging
logger = logging.getLogger(__name__)

import os
from contextlib import contextmanager

import gi
gi.require_version('Gtk', '3.0')  # @UndefinedVariable
from gi.repository import GObject, Pango, Gdk # @UnresolvedImport

from mvc import Controller
from mvc.adapters.gtk_support.dialogs.dialog_factory import DialogFactory
from mvc.adapters.gtk_support.widgets.threaded_task_box import ThreadedTaskBox
from mvc.support.cancellable_thread import CancellableThread
from mvc.support.gui_loop import run_when_idle

from pyxrd.generic.controllers.line_controllers import BackgroundController
from pyxrd.generic.views.line_views import BackgroundView
from pyxrd.generic.views.treeview_tools import new_text_column, new_toggle_column, new_pb_column
from pyxrd.generic.controllers import BaseController, ObjectListStoreController

from pyxrd.file_parsers.xrd_parsers import xrd_parsers

from pyxrd.specimen.models import Specimen



class ProjectController(ObjectListStoreController):

    treemodel_property_name = "specimens"
    treemodel_class_type = Specimen
    columns = [ ]
    delete_msg = "Deleting a specimen is irreversible!\nAre You sure you want to continue?"
    auto_adapt = True

    def register_view(self, view):
        if view is not None and self.model is not None:
            if self.parent is not None: # is this still needed?
                tv = self.view["project_specimens"]
                tv.set_model(self.treemodel)
                self.view.treeview = tv
                self.view.set_x_range_sensitive(self.model.axes_xlimit == 1)
                self.view.set_y_range_sensitive(self.model.axes_ylimit == 1)
        return

    def _idle_register_view(self, *args, **kwargs):
        super(ProjectController, self)._idle_register_view(*args, **kwargs)

    def adapt(self, *args, **kwargs):
        super(ProjectController, self).adapt(*args, **kwargs)

    def setup_treeview(self, widget):
        super(ProjectController, self).setup_treeview(widget)
        store = self.treemodel
        widget.connect('button-press-event', self.specimen_tv_button_press)

        # First reset & then (re)create the columns of the treeview:
        for col in widget.get_columns():
            widget.remove_column(col)

        # Name column:
        col = new_text_column('Name',
            text_col=store.c_name,
            min_width=125,
            xalign=0.0,
            ellipsize=Pango.EllipsizeMode.END)
        setattr(col, "colnr", store.c_name)
        widget.append_column(col)

        # Check boxes:
        def toggle_renderer(column, cell, model, itr, data=None):
            active = False
            if model.iter_is_valid(itr):
                col = column.get_col_attr("active")
                active = model.get_value(itr, col)
            cell.set_property('active', active)
            return
        def setup_check_column(title, colnr):
            col = new_toggle_column(title,
                    toggled_callback=(self.specimen_tv_toggled, (store, colnr)),
                    data_func=toggle_renderer,
                    resizable=False,
                    expand=False,
                    activatable=True,
                    active_col=colnr)
            setattr(col, "colnr", colnr)
            widget.append_column(col)

        setup_check_column('Exp', store.c_display_experimental)
        if self.model.layout_mode == "FULL":
            setup_check_column('Cal', store.c_display_calculated)
            setup_check_column('Sep', store.c_display_phases)

        # Up and down arrows:
        def setup_image_button(image, colnr):
            col = new_pb_column("", resizable=False, expand=False, stock_id=image)
            setattr(col, "colnr", colnr)
            widget.append_column(col)
        setup_image_button("213-up-arrow", 501)
        setup_image_button("212-down-arrow", 502)

    def edit_object(self, obj):
        pass # clear this method, we're not having an 'edit' view pane...

    @BaseController.status_message("Importing multiple specimens...", "add_specimen")
    def import_multiple_specimen(self):
        def on_accept(dialog):
            ## TODO MOVE THIS (PARTIALLY?) TO THE MODEL LEVEL ##

            filenames = dialog.get_filenames()
            parser = getattr(dialog.get_filter(), "parser")

            task = ThreadedTaskBox()
            window = DialogFactory.get_custom_dialog(
                 task, parent=self.view.get_top_widget())

            # Status:
            status_dict = dict(
                total_files=len(filenames),
                current_file=0,
                specimens=[]
            )

            # Task:
            def load_specimens(stop=None):
                for filename in filenames:
                    if stop is not None and stop.is_set():
                        return
                    # Error message is case parsing fails:
                    message = "An unexpected error has occurred when trying to parse %s:\n\n<i>" % os.path.basename(filename)
                    message += "%s</i>\n\n"
                    message += "This is most likely caused by an invalid or unsupported file format."
                    # Run & report any errors:
                    with DialogFactory.error_dialog_handler(
                            message, self.view.get_top_widget(), 
                            title="Failed to load file", reraise=False):
                        specimens = Specimen.from_experimental_data(filename=filename, parent=self.model, parser=parser)
                        status_dict["specimens"] += specimens
                    status_dict["current_file"] += 1

            # Cancel & stop events:
            def on_interrupted(*args, **kwargs):
                window.hide()

            # Status label update:
            def gui_callback():
                task.set_status("Loading file %d/%d ..." % (
                    status_dict["current_file"],
                    status_dict["total_files"]
                ))
                return True
            gui_timeout_id = GObject.timeout_add(250, gui_callback)

            # Complete event:
            @run_when_idle
            def on_complete(*args, **kwargs):
                last_iter = None
                for specimen in status_dict["specimens"]:
                    last_iter = self.model.specimens.append(specimen)
                if last_iter is not None:
                    self.view["project_specimens"].set_cursor(last_iter)
                GObject.source_remove(gui_timeout_id)
                window.hide()
                window.destroy()

            # Run task box:
            task.connect("cancelrequested", on_interrupted)
            task.connect("stoprequested", on_interrupted)
            task.set_status("Loading ...")
            task.start()
            window.show_all()

            # Run thread:
            self.thread = CancellableThread(load_specimens, on_complete)
            self.thread.start()

        DialogFactory.get_load_dialog(title="Select XRD files for import",
                             filters=xrd_parsers.get_import_file_filters(),
                             parent=self.view.get_top_widget(),
                             multiple=True).run(on_accept)

    @BaseController.status_message("Deleting specimen...", "del_specimen")
    def delete_selected_specimens(self):
        """
            Asks the user for confirmation and if positive deletes all the 
            selected specimens. Does nothing when no specimens are selected.
        """
        selection = self.get_selected_objects()
        if selection is not None and len(selection) >= 1:
            def delete_objects(dialog):
                for obj in selection:
                    if obj is not None:
                        self.model.specimens.remove(obj)
            DialogFactory.get_confirmation_dialog(
                message='Deleting a specimen is irreversible!\nAre You sure you want to continue?',
                parent=self.view.get_top_widget()
            ).run(delete_objects)

    @BaseController.status_message("Removing backgrounds...", "del_bg_specimen")
    def remove_backgrounds(self, specimens):
        """
            Opens the 'remove background' dialog for the given specimens,
            raises a ValueError error if (one of) the specimens is not part of
            this project.
        """

        def on_automated(dialog):
            for specimen in specimens:
                if not specimen in self.model.specimens:
                    raise ValueError("Specimen `%s` is not part of this Project!" % specimen)
                else:
                    specimen.experimental_pattern.bg_type = 0 # Linear see settings
                    specimen.experimental_pattern.find_bg_position()
                    specimen.experimental_pattern.remove_background()
                    specimen.experimental_pattern.clear_bg_variables()

        def on_not_automated(dialog):
            for specimen in specimens:
                if not specimen in self.model.specimens:
                    raise ValueError("Specimen `%s` is not part of this Project!" % specimen)
                else:
                    bg_view = BackgroundView(parent=self.parent.view)
                    BackgroundController(model=specimen.experimental_pattern, view=bg_view, parent=self)
                    bg_view.present()

        # Ask user if he/she wants automation:
        DialogFactory.get_confirmation_dialog(
            "Do you want to perform an automated linear background subtraction?",
            parent=self.parent.view.get_top_widget()
        ).run(on_automated, on_not_automated)

    def edit_specimen(self):
        selection = self.get_selected_objects()
        if selection is not None and len(selection) == 1:
            # TODO move the specimen view & controller into the project level
            self.parent.view.specimen.present()

    @BaseController.status_message("Creating new specimen...", "add_specimen")
    def add_specimen(self):
        specimen = Specimen(parent=self.model, name="New Specimen")
        self.model.specimens.append(specimen)
        self.view.specimens_treeview.set_cursor(self.treemodel.on_get_path(specimen))
        self.edit_specimen()
        return True

    @contextmanager
    def _multi_operation_context(self):
        with self.model.hold_mixtures_data_changed():
            with self.model.data_changed.hold():
                yield

    # ------------------------------------------------------------
    #      Notifications of observable properties
    # ------------------------------------------------------------
    @Controller.observe("name", assign=True)
    def notif_change_name(self, model, prop_name, info):
        self.parent.update_title()
        return

    @Controller.observe("axes_xlimit", assign=True)
    def notif_xlimit_toggled(self, model, prop_name, info):
        self.view.set_x_range_sensitive(int(self.model.axes_xlimit) == 1)

    @Controller.observe("axes_ylimit", assign=True)
    def notif_ylimit_toggled(self, model, prop_name, info):
        self.view.set_y_range_sensitive(int(self.model.axes_ylimit) == 1)

    @Controller.observe("layout_mode", assign=True)
    def notif_layout_mode(self, model, prop_name, info):
        self.parent.set_layout_mode(self.model.layout_mode)
        if self.view is not None:
            self.setup_treeview(self.view.treeview)

    # ------------------------------------------------------------
    #      GTK Signal handlers
    # ------------------------------------------------------------
    def specimen_tv_toggled(self, cell, path, model, colnr):
        if model is not None:
            itr = model.get_iter(path)
            model.set_value(itr, colnr, not cell.get_active())
            return True
        return False

    def specimen_tv_button_press(self, tv, event):
        specimen = None
        current_specimens = self.parent.model.current_specimens or []
        ret = tv.get_path_at_pos(int(event.x), int(event.y))
        if ret is not None:
            path, col, x, y = ret
            specimen = self.treemodel.get_user_data_from_path(path)
        if event.button == 3:
            if specimen is not None:
                # clicked a specimen which is not in the current selection,
                # so clear selection and select it
                if not specimen in current_specimens:
                    self.select_object(specimen)
            else:
                # clicked an empty space, so clear selection
                self.select_object(None)
            self.view.show_specimens_context_menu(event)
            return True
        elif event.type == Gdk.EventType._2BUTTON_PRESS and specimen is not None and getattr(col, "colnr") == self.treemodel.c_name:  # @UndefinedVariable
            self.parent.on_edit_specimen_activate(event)
            return True
        elif (event.button == 1 or event.type == Gdk.EventType._2BUTTON_PRESS) and specimen is not None:  # @UndefinedVariable
            column = getattr(col, "colnr")
            if column in (self.treemodel.c_display_experimental,
                    self.treemodel.c_display_calculated,
                    self.treemodel.c_display_phases):
                if column == self.treemodel.c_display_experimental:
                    specimen.display_experimental = not specimen.display_experimental
                elif column == self.treemodel.c_display_calculated:
                    specimen.display_calculated = not specimen.display_calculated
                elif column == self.treemodel.c_display_phases:
                    specimen.display_phases = not specimen.display_phases
                # TODO FIXME self.treemodel.on_row_changed(ret)
                return True
            elif column == 501:
                self.model.move_specimen_down(specimen)
                self.parent.model.current_specimens = self.get_selected_objects()
                return True
            elif column == 502:
                self.model.move_specimen_up(specimen)
                self.parent.model.current_specimens = self.get_selected_objects()
                return True

    def objects_tv_selection_changed(self, selection):
        ObjectListStoreController.objects_tv_selection_changed(self, selection)
        self.parent.model.current_specimens = self.get_selected_objects()
        return True

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on
#  -------------------------------------------------------------------------
#  Copyright (C) 2014 by Mathijs Dumon <mathijs dot dumon at gmail dot com>
#
#  mvc is a framework derived from the original pygtkmvc framework
#  hosted at: <http://sourceforge.net/projects/pygtkmvc/>
#
#  mvc is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2 of the License, or (at your option) any later version.
#
#  mvc is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110, USA.
#  -------------------------------------------------------------------------

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk

from .basic import GtkAdapter

class LabelAdapter(GtkAdapter):
    """
        An adapter for a Gtk.Label widget
    """
    widget_types = ["label"]
    _check_widget_type = Gtk.Label

    _wid_read = lambda c, w: Gtk.Label.get_text(w)
    _wid_write = lambda c, w, v: Gtk.Label.set_text(w, str(v))

    _signal = None

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import logging
logger = logging.getLogger(__name__)

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk
import sys

from mvc.adapters.gtk_support.dialogs.dialog_factory import DialogFactory
from mvc.support.cancellable_thread import CancellableThread
from mvc.support.gui_loop import run_when_idle, add_timeout_call,\
    remove_timeout_call

from pyxrd.generic.asynchronous.providers import get_status
from pyxrd.generic.views.treeview_tools import new_text_column, new_pb_column, new_toggle_column
from pyxrd.generic.mathtext_support import create_pb_from_mathtext
from pyxrd.generic.controllers import DialogController

from pyxrd.refinement.views.refiner_view import RefinerView
from pyxrd.refinement.controllers.refiner_controller import RefinerController

class RefinementController(DialogController):

    auto_adapt_included = [
        "refine_method_index",
        "refinables",
        "make_psp_plots",
    ]

    @property
    def treemodel(self):
        return self.model.refinables

    def setup_refinables_tree_view(self, store, widget):
        """
            Setup refinables TreeView layout
        """
        widget.set_show_expanders(True)

        if sys.platform == "win32":
            def get_label(column, cell, model, itr, user_data=None):
                ref_prop = model.get_tree_node_object(itr)
                cell.set_property("text", ref_prop.text_title)
                return
            widget.append_column(new_text_column('Name/Prop', xalign=0.0, data_func=get_label))
        else:
            # Labels are parsed for mathtext markup into pb's:
            def get_pb(column, cell, model, itr, user_data=None):
                ref_prop = model.get_tree_node_object(itr)
                try:
                    if not hasattr(ref_prop, "pb") or not ref_prop.pb:
                        ref_prop.pb = create_pb_from_mathtext(
                            ref_prop.title,
                            align='left',
                            weight='medium'
                        )
                    cell.set_property("pixbuf", ref_prop.pb)
                except RuntimeError:
                    logger.warning("An error occured when trying to convert a property title to a PixBuf")
                    raise
                return
            widget.append_column(new_pb_column('Name/Prop', xalign=0.0, data_func=get_pb))

        # Editable floats:
        def get_value(column, cell, model, itr, *args):
            col = column.get_col_attr('markup')
            try:
                value = model.get_value(itr, col)
                value = "%.5f" % value
            except TypeError: value = ""
            cell.set_property("markup", value)
            return
        def on_float_edited(rend, path, new_text, model, col):
            itr = model.get_iter(path)
            try:
                model.set_value(itr, col, float(new_text))
            except ValueError:
                return False
            return True

        def_float_args = {
            "sensitive_col": store.c_refinable,
            "editable_col": store.c_refinable,
            "visible_col": store.c_refinable,
            "data_func": get_value
        }

        widget.append_column(new_text_column(
            "Value", markup_col=store.c_value,
            edited_callback=(
                on_float_edited,
                (store, store.c_value,)
            ), **def_float_args
        ))
        widget.append_column(new_text_column(
            "Min", markup_col=store.c_value_min,
            edited_callback=(
                on_float_edited,
                (store, store.c_value_min,)
            ), **def_float_args
        ))
        widget.append_column(new_text_column(
            "Max", markup_col=store.c_value_max,
            edited_callback=(
                on_float_edited,
                (store, store.c_value_max,)
            ), **def_float_args
        ))

        # The 'refine' checkbox:
        widget.append_column(new_toggle_column(
            "Refine",
            toggled_callback=(self.refine_toggled, (store,)),
            resizable=False,
            expand=False,
            active_col=store.c_refine,
            sensitive_col=store.c_refinable,
            activatable_col=store.c_refinable,
            visible_col=store.c_refinable
        ))

    def _update_method_options_store(self):
        """
            Update the method options tree store (when a new method is selected)
        """
        tv = self.view['tv_method_options']
        store = Gtk.ListStore(str, str)
        method = self.model.get_refinement_method()
        for arg in method.options:
            description = getattr(type(method), arg).description
            store.append([arg, description])
        tv.set_model(store)
        return tv

    def _setup_method_options_treeview(self):
        """
            Initial method options tree view layout & behavior setup
        """
        # Update the method options store to match the currently selected
        # refinement method
        tv = self._update_method_options_store()

        # The name of the option:
        tv.append_column(new_text_column("Name", text_col=1))

        # The value of the option:
        def get_value(column, cell, model, itr, *args):
            option_name, = tv.get_model().get(itr, 0)
            method = self.model.get_refinement_method()
            cell.set_property("sensitive", True)
            cell.set_property("editable", True)
            cell.set_property("markup", "%g" % getattr(method, option_name))
            return
        
        def on_value_edited(rend, path, new_text, col):
            store = tv.get_model()
            itr = store.get_iter(path)
            option_name, = store.get(itr, 0)
            method = self.model.get_refinement_method()
            try:
                setattr(method, option_name, new_text)
            except ValueError:
                pass
            return True
        tv.append_column(new_text_column(
            "Value", text_col=0,
            data_func=get_value,
            edited_callback=(on_value_edited, (0,)),
        ))

    def register_view(self, view):
        # Create the method treeview:
        self._setup_method_options_treeview()
        # Update the server status:
        self.view.update_server_status(get_status())

    def cleanup(self):
        if hasattr(self, "view"):
            del self.view
        if hasattr(self, "results_view"):
            del self.results_view
        if hasattr(self, "results_controller"):
            del self.results_controller
        if hasattr(self, "model"):
            self.relieve_model(self.model)

    # ------------------------------------------------------------
    #      Notifications of observable properties
    # ------------------------------------------------------------
    @DialogController.observe("refine_method_index", assign=True)
    def on_prop_changed(self, model, prop_name, info):
        self._update_method_options_store()

    # ------------------------------------------------------------
    #      GTK Signal handlers
    # ------------------------------------------------------------
    def on_cancel(self):
        if self.view is not None:
            self.view.hide()
            self.parent.view.parent.show()

    def refine_toggled(self, cell, path, model):
        if model is not None:
            itr = model.get_iter(path)
            model.set_value(itr, model.c_refine, not cell.get_active())
        return True

    def on_btn_randomize_clicked(self, event):
        self.model.randomize()

    def on_auto_restrict_clicked(self, event):
        self.model.auto_restrict()

    def _launch_gui_updater(self, refiner):
        def _on_update_gui():
            if self.view is not None and refiner is not None:
                self.view.update_refinement_info(
                    refiner.history.last_residual,
                    refiner.status.message,
                    get_status()
                )
                return True
            else:
                return False
        add_timeout_call(500, _on_update_gui)
        return _on_update_gui

    def _launch_refine_thread(self, refiner, gui_timeout_id):
        @run_when_idle
        def thread_completed(*args, **kwargs):
            """ Called when the refinement is completed """
            self.thread = None
            
            remove_timeout_call(gui_timeout_id)
            self.view.stop_spinner()
            
            # Make some plots:
            if self.model.make_psp_plots:
                self.view.update_refinement_status("Processing...")
                self.results_controller.generate_images()
            
            # Set the labels:
            self.results_controller.update_labels()
                
            # Hide our shit:
            self.view.hide_refinement_info()
            self.view.hide()
            
            # Show results:
            self.results_view.present()

        thread = CancellableThread(refiner.refine, thread_completed)
        thread.start()
        return thread

    def _connect_cancel_button(self, refiner, gui_timeout_id, thread):
        # Connect the cancel button (custom widget):
        def thread_cancelled(*args, **kwargs):
            """ Called when the refinement is cancelled by the user """
            remove_timeout_call(gui_timeout_id)
            self.view.stop_spinner()
            
            self.view.update_refinement_status("Cancelling...")
            thread.cancel()
            
            self.view.hide_refinement_info()
        self.view.connect_cancel_request(thread_cancelled)

    @DialogController.status_message("Refining mixture...", "refine_mixture")
    def on_refine_clicked(self, event):
        with self.model.mixture.needs_update.hold():
            with self.model.mixture.data_changed.hold():
                if len(self.model.mixture.specimens) > 0:
                    # Create the refiner object
                    with DialogFactory.error_dialog_handler(
                            "There was an error when creating the refinement setup:\n{}", 
                            parent=self.view.get_toplevel(), reraise=False):
                        refiner = self.model.get_refiner()

                        # Setup results controller
                        self.results_view = RefinerView(parent=self.view.parent)
                        self.results_controller = RefinerController(
                            refiner=refiner,
                            model=self.model,
                            view=self.results_view,
                            parent=self
                        )
    
                        # Gtk timeout loop for our GUI updating:
                        gui_timeout_id = self._launch_gui_updater(refiner)
                        
                        # This creates a thread that will run the refiner.refine method:
                        thread = self._launch_refine_thread(refiner, gui_timeout_id)
                        
                        # Connect the cancel button:
                        self._connect_cancel_button(refiner, gui_timeout_id, thread)
    
                        # Show the context updates in the gui:
                        self.view.show_refinement_info()
                        self.view.start_spinner()

                else:
                    DialogFactory.get_information_dialog(
                        "Cannot refine an empty mixture!", parent=self.view.get_toplevel()
                    ).run()

    pass # end of class


# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

"""
    The following classes are not meant to be used directly, rather you should
    create the corresponding model instances and retrieve the DataObject from 
    them.
    
    The rationale behind not using the model instances directly is that
    they are difficult to serialize or pickle (memory-)efficiently.
    This is mainly due to all of the boiler-plate code that takes care of
    references, saving, loading, calculating properties from other properties
    etc. A lot of this is not needed for the actual calculation.    
    The data objects below, on the other hand, only contain the data needed to
    be able to calculate XRD patterns.
"""

class DataObject(object):
    """
    The base class for all DataObject instances.
    
    The constructor takes any number of keyword arguments it will set as
    attributes on the instance.
    """
    def __init__(self, **kwargs):
        for key, val in kwargs.items():
            setattr(self, key, val)

    pass # end of class

class AtomTypeData(DataObject):
    """ The DataObject describing an AtomType. """

    #: a numpy array of `a` scattering factors
    par_a = None
    #: a numpy array of `b` scattering factors
    par_b = None
    #: the `c` scattering constant
    par_c = None
    #: the debye-waller temperature factor
    debye = None

    pass # end of class

class AtomData(DataObject):
    """ The DataObject describing an Atom. """

    #: an :class:`~AtomTypeData` instance
    atom_type = None
    #: the # of atoms projected to this z coordinate
    pn = None
    #: the default z coordinate
    default_z = None
    #: the actual z coordinate
    z = None

    pass # end of class

class ComponentData(DataObject):
    """ The DataObject describing an Atom """

    #: a list of :class:`~AtomData` instances
    layer_atoms = None

    #: a list of :class:`~AtomData` instances
    interlayer_atoms = None

    #: the component volume
    volume = None

    #: the component weight
    weight = None

    #: the d-spacing of the component
    d001 = None

    #: the default d-spacing of the component
    default_c = None

    #: the variation in d-spacing of the component
    delta_c = None

    #: the height of the silicate lattice (excluding the interlayer space)
    lattice_d = None

    pass # end of class

class CSDSData(DataObject):
    """ The DataObject describing the CSDS distribution. """

    #: average CSDS
    average = None

    #: maximum CSDS
    maximum = None

    #: minimum CSDS
    minimum = None

    #: the alpha scale factor for the log-normal distribution
    alpha_scale = None

    #: the alpha offset factor for the log-normal distribution
    alpha_offset = None

    #: the beta scale factor for the log-normal distribution
    beta_scale = None

    #: the beta offset factor for the log-normal distribution
    beta_offset = None

    pass # end of class

class GonioData(DataObject):
    """ The DataObject describing the Goniometer setup. """

    #: Lower 2-theta bound for calculated patterns
    min_2theta = None

    #: Upper 2-theta bound for calculated patterns
    max_2theta = None

    #: The number of steps in between the lower and upper 2-theta bounds
    steps = None

    #: If the first soller slits are present
    has_soller1 = False
    
    #: The first soller slit size
    soller1 = None

    #: If the first soller slits are present
    has_soller2 = False

    #: The second soller slit size
    soller2 = None

    #: The divergence slit mode
    divergence_mode = "FIXED"

    #: The divergence size (degrees (fixed) or mm (auto))
    divergence = None

    #: The Bragg angle of the monochromator (or 0° if not present)
    mcr_2theta = 0

    #: Flag indicating if intensities need to be corrected for absorption 
    has_absorption_correction = None
    
    #: The sample mass absorption coefficient (mg/cm²)
    absorption = 45.0
    
    #: The sample surface density (cm²/g)
    sample_surf_density = 20.0

    #: The goniometer radius
    radius = None

    #: The goniometer wavelength
    wavelength = None

    #: The goniometer wavelength distribution
    wavelength_distribution = None
    
    #: The sample length
    sample_length = None

    pass # end of class

class ProbabilityData(DataObject):
    """ The DataObject describing the layer stacking probabilities """

    #: Whether this probability is really a valid one
    valid = None

    #: The number of components this probability describes
    G = None

    #: The weight fractions matrix
    W = None

    #: The probabilities matrix
    P = None

    pass # end of class

class PhaseData(DataObject):
    """ The DataObject describing a phase """

    #: A flag indicating whether to apply Lorentz-polarization factor or not
    apply_lpf = True

    #: A flag indicating whether to apply machine corrections or not
    apply_correction = True

    #: A list of :class:`~ComponentData` instances
    components = None

    #: A :class:`~ProbabilityData` instance
    probability = None

    #: The sigma start value
    sigma_star = None

    #: A :class:`~CSDSData` instance
    csds = None

    pass # end of class

class SpecimenData(DataObject):
    """ The DataObject describing a specimen """

    #: A :class:`~GonioData` instance
    goniometer = None

    #: The sample absorption
    absorption = None

    #: A list of :class:`~PhaseData` instances
    phases = None

    #: A numpy array with the observed intensities
    observed_intensity = None

    #: A numpy array with the calculated intensity
    total_intensity = None

    #: A nummpy array with the calculated phase profiles
    phase_intensities = None

    #: A numpy array with a correction factor taking the sample & goniometer
    #  properties into account
    correction = None

    pass # end of class

class MixtureData(DataObject):
    """ The DataObject describing a mixture """

    #: A list of :class:`~SpecimenData` instances
    specimens = None

    #: A numpy array with the phase fractions
    fractions = None

    #: A numpy array with the specimen background shifts
    bgshifts = None

    #: A numpy array with the specimen absolute scales
    scales = None

    #: Whether this MixtureData object has been parsed (internal flag)
    parsed = False
    
    #: Whether this MixtureData object has been calculated (internal flag)
    calculated = False

    #: Whether this MixtureData object has been optimized (internal flag)
    optimized = False

    #: The number of specimens
    n = 0

    #: The number of phases
    m = 0

    pass # end of class



#!/usr/bin/python

# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import gi
gi.require_version('Gtk', '3.0')  # @UndefinedVariable
from gi.repository import GObject  # @UnresolvedImport

import unittest

from pyxrd.generic.models import DataModel

from mvc.adapters.gtk_support.treemodels import ObjectListStore
from mvc.models.properties import (
    LabeledProperty, StringProperty, FloatProperty
)

__all__ = [
    'TestObjectListStore',
]


class _DummyObject(DataModel):

    name = StringProperty(text="Name", tabular=True, default="")
    number = FloatProperty(text="Number", tabular=True, default=0)
    test = LabeledProperty(text="Test", tabular=True, default=[])

    pass # end of class

class _DummyParent(DataModel):

    attrib = LabeledProperty(text="Attrib", default=[], tabular=True, data_type=_DummyObject)

    pass # end of class


class TestObjectListStore(unittest.TestCase):

    def setUp(self):
        self.model = _DummyParent()
        prop = type(self.model).attrib
        self.store = ObjectListStore(self.model, prop)

    def tearDown(self):
        super(TestObjectListStore, self).tearDown()
        del self.model

    def test_columns(self):
        self.assertEqual(self.store.get_n_columns(), len(_DummyObject.Meta.get_column_properties()))
        self.assertEqual(self.store.get_column_type(self.store.c_name), GObject.type_from_name("gchararray"))
        self.assertEqual(self.store.get_column_type(self.store.c_number), GObject.type_from_name("gdouble"))
        self.assertEqual(self.store.get_column_type(self.store.c_test), GObject.type_from_name("PyObject"))

    def test_convert(self):
        self.assertEqual(self.store.convert(self.store.c_number, "0.5"), 0.5)

    # TODO:
    # - test JSON serialisation
    # - test raw data

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from .cell_renderer_tools import get_default_renderer, parse_callback, parse_kwargs

def add_renderer_with_attrs(combo, col_attrs, rend):
    combo.pack_start(rend, True)
    for attr, val in col_attrs.items():
        combo.add_attribute(rend, attr, val)

def add_combo_text_column(combo,
        data_func=None,
        **kwargs):
    kwargs["xalign"] = kwargs.get("xalign", 0.0)
    kwargs, col_attrs = parse_kwargs(**kwargs)
    rend = get_default_renderer('text', **kwargs)
    add_renderer_with_attrs(combo, col_attrs, rend)
    if data_func!=None:
        callback, args = parse_callback(data_func)
        combo.set_cell_data_func(rend, callback, args)
    return rend



#!/usr/bin/python
# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import unittest

from test.test_generic.test_io.test_file_parsers import BaseTestParsers, load_data_from_files

from pyxrd.file_parsers.xrd_parsers import BrkRAWParser


__all__ = [
    'TestRAWParser',
]

class TestRAWParser(BaseTestParsers.BaseTestParser):

    parser_class = BrkRAWParser
    file_data = load_data_from_files(
        "test/test_generic/test_io/test_xrd_parsers/brk_raw1.raw",
        "test/test_generic/test_io/test_xrd_parsers/brk_raw2.raw"
    )

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from pyxrd.generic.io import get_case_insensitive_glob

from pyxrd.project.importing import create_project_from_sybilla_xml

from ..base_parser import BaseParser
from ..xml_parser_mixin import XMLParserMixin
from .namespace import project_parsers

@project_parsers.register_parser()
class SybillaProjectParser(XMLParserMixin, BaseParser):

    description = "Sybilla XML files"
    extensions = get_case_insensitive_glob("*.XML")
    mimetypes = ["application/xml", "text/xml"]

    @classmethod
    def parse(cls, fp, data_objects=None, close=True):
        """
            This method parses the file and return a list of DataObjects
            with both header and data properties filled in accordingly.
            The filename argument is always required. If no file object is passed
            as keyword argument, it only serves as a label. Otherwise a new file
            object is created.
            File objects are closed unless close is set to False.
            Existing DataObjects can be passed as well and will then 
            be used instead of creating new ones.
        """
        filename, fp, close = cls._get_file(fp, close=close)
        data_objects = create_project_from_sybilla_xml(filename)
        if close: fp.close()
        return data_objects

    pass #end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

raise NotImplemented("This module is not yet implemented") 

import sys, imp

from contextlib import contextmanager

import numpy as np

from pyxrd.generic.io.custom_io import storables
from pyxrd.mixture.models.mixture import Mixture
from copy import deepcopy

@storables.register()
class InSituMixture(Mixture):
    """
        Advanced model for optimizing and refining of calculated in-situ XRD data.
        Is a sub-class of the basic Mixture class.
        
        Each phase is linked with an additional Behaviour instance - these are created at the
        project level.
    """
    
    # MODEL INTEL:
    class Meta(Mixture.Meta):
        store_id = "InSituMixture"
    
    _data_object = None
    @property
    def data_object(self):
        self._data_object = Mixture.data_object.fget(self)  # @UndefinedVariable
        return self._data_object
    
    def get_phase_data_object(self, specimen_index, z_index, phase_index):
        behav = self.behaviour_matrix[specimen_index, ...].flatten()[phase_index] if self.behaviour_matrix is not None else None
        phase = self.phase_matrix[specimen_index, ...].flatten()[phase_index] if self.phase_matrix is not None else None
        if behav is not None and phase is not None:
            behav.apply(phase, self.specimens[specimen_index].get_z_list()[z_index])
        return deepcopy(phase.data_object) if phase is not None else None
    
    # Lists and matrices:
    
    #: A 2D numpy object array containing the behaviour paths
    behaviour_matrix = None 
    
    # ------------------------------------------------------------
    #      Initialization and other internals
    # ------------------------------------------------------------
    def __init__(self, *args, **kwargs):

        my_kwargs = self.pop_kwargs(kwargs,
            "behaviour_uuids",
            *[prop.label for prop in InSituMixture.Meta.get_local_persistent_properties()]
        )
        super(InSituMixture, self).__init__(*args, **kwargs)
        kwargs = my_kwargs
        
        with self.data_changed.hold():
            # 2D matrix, rows match specimens, columns match mixture 'phase behaviours'; contains the actual behaviour objects           
            behaviour_uuids = self.get_kwarg(kwargs, None, "behaviour_uuids")
            if behaviour_uuids is not None:
                self.behaviour_matrix = np.array([[type(type(self)).object_pool.get_object(uuid) if uuid else None for uuid in row] for row in behaviour_uuids], dtype=np.object_)
            else:
                self.behaviour_matrix = np.empty(shape=(0, 0), dtype=np.object_)
        
            print(behaviour_uuids, self.behaviour_matrix)
        
            pass # end hold data_changed
        
    # ------------------------------------------------------------
    #      Input/Output stuff
    # ------------------------------------------------------------
    def json_properties(self):
        retval = super(InSituMixture, self).json_properties()
        retval["behaviour_uuids"] = [[item.uuid if item else None for item in row] for row in map(list, self.behaviour_matrix)]
        return retval
        
    @staticmethod
    def from_json(**kwargs):
        # Remove this deprecated kwarg:
        if "refinables" in kwargs:
            del kwargs["refinables"]
        return InSituMixture(**kwargs)
        
    # ------------------------------------------------------------
    #      Methods & Functions
    # ------------------------------------------------------------
       
    def get_behaviour(self, specimen_slot, phase_slot):
        """Returns the behaviour at the given slot positions or None if not set"""
        return self.behaviour_matrix[specimen_slot, phase_slot]

    def set_behaviour(self, specimen_slot, phase_slot, behaviour):
        """Sets the behaviour at the given slot positions"""
        if self.parent is not None: #no parent = no valid phases
            with self.needs_update.hold_and_emit():
                with self.data_changed.hold():
                    if behaviour is None:
                        raise RuntimeError("Behaviour can not be set to None, use unset_behaviour to clear!")
                    self.behaviour_matrix[specimen_slot, phase_slot] = behaviour
                    self.refinement.update_refinement_treestore()
                    
    def unset_behaviour(self, behaviour):
        """ Clears a behaviour slot in the phase matrix """
        with self.needs_update.hold_and_emit():
            with self.data_changed.hold():
                shape = self.behaviour_matrix.shape
                for i in range(shape[0]):
                    for j in range(shape[1]):
                        if self.behaviour_matrix[i, j] == behaviour:
                            self.behaviour_matrix[i, j] = None
                self.refinement.update_refinement_treestore()
    
    @contextmanager
    def _relieve_and_observe_behaviours(self):
        self._relieve_behaviours()
        yield
        self._observe_behaviours()

    def _observe_behaviours(self):
        """ Starts observing behaviour in the behaviour matrix"""
        for behaviour in self.behaviour_matrix.flat:
            if behaviour is not None:
                self.observe_model(behaviour)

    def _relieve_behaviours(self):
        """ Relieves behaviour observer calls """
        for behaviour in self.behaviour_matrix.flat:
            if behaviour is not None:
                self.relieve_model(behaviour)
    
    def add_phase_slot(self, phase_name, fraction):
        """ Adds a new phase column to the phase matrix """
        with self.needs_update.hold_and_emit():
            with self.data_changed.hold():
                n, m = self.behaviour_matrix.shape if self.behaviour_matrix.ndim == 2 else (0, 0)
                if self.behaviour_matrix.size == 0:
                    self.behaviour_matrix = np.resize(self.behaviour_matrix.copy(), (n, m + 1))
                    self.behaviour_matrix[:] = None
                else:
                    self.behaviour_matrix = np.concatenate([self.behaviour_matrix.copy(), [[None]] * n ], axis=1)
                    self.behaviour_matrix[:, m] = None
                return super(InSituMixture, self).add_phase_slot(phase_name, fraction)
                
    def del_phase_slot(self, phase_slot):
        """ Deletes a phase column using its index """
        with self.needs_update.hold_and_emit():
            with self.data_changed.hold():
                with self._relieve_and_observe_behaviours():
                    self.behaviour_matrix = np.delete(self.behaviour_matrix, phase_slot, axis=1)
                return super(InSituMixture, self).del_phase_slot(phase_slot)

    def add_specimen_slot(self, specimen, scale, bgs):
        """ Adds a new specimen to the phase matrix (a row) and specimen list """
        with self.needs_update.hold_and_emit():
            with self.data_changed.hold():                   
                n, m = self.behaviour_matrix.shape if self.behaviour_matrix.ndim == 2 else (0, 0)
                if self.behaviour_matrix.size == 0:
                    self.behaviour_matrix = np.resize(self.behaviour_matrix.copy(), (n + 1, m))
                    self.behaviour_matrix[:] = None
                else:
                    self.behaviour_matrix = np.concatenate([self.behaviour_matrix.copy(), [[None] * m] ], axis=0)
                    self.behaviour_matrix[n, :] = None
                return super(InSituMixture, self).add_specimen_slot(specimen, scale, bgs)           
    
    def del_specimen_slot(self, specimen_slot):
        """ Deletes a specimen slot using its slot index """
        with self.needs_update.hold_and_emit():
            with self.data_changed.hold():
                self.behaviour_matrix = np.delete(self.behaviour_matrix, specimen_slot, axis=0)
                return super(InSituMixture, self).del_specimen_slot(specimen_slot)
               
    # ------------------------------------------------------------
    #      Refinement stuff:
    # ------------------------------------------------------------
    def set_data_object(self, mixture, calculate=False):
        """
            Sets the fractions, scales and bgshifts of this mixture.
        """
        if mixture is not None:
            with self.needs_update.ignore():
                with self.data_changed.hold_and_emit():
                    self.fractions[:] = list(mixture.fractions)
                    self.scales[:] = list(mixture.scales)
                    self.bgshifts[:] = list(mixture.bgshifts)

                    if calculate: # (re-)calculate if requested:
                        mixture = self.optimizer.calculate(mixture)

                    for i, (specimen_data, specimen) in enumerate(zip(mixture.specimens, self.specimens)):
                        if specimen is not None:
                            with specimen.data_changed.ignore():
                                specimen.update_pattern(
                                    specimen_data.total_intensity,
                                    specimen_data.phase_intensities * self.fractions[:, np.newaxis] * self.scales[i],
                                    self.phase_matrix[i, :]
                                )
                    
                   
    def load_behaviour_from_module(self, path):
        """Loads the behaviour class from the given filepath"""
        mod = imp.load_source('Behaviour', path)
        del sys.modules['Behaviour'] # move the module to a better spot
        sys.modules[mod.BEHAVIOUR_CLASS] = mod
        klass = mod[mod.BEHAVIOUR_CLASS]
        
        behaviour = klass()
        behaviour.__path = path # FIXME 
        
        return behaviour
        
    pass # end of class


#!/usr/bin/python

# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import time, atexit, os
from traceback import print_exc

import logging
logger = logging.getLogger(__name__)

import Pyro4
import Pyro4.naming

try:
    import threading as threading
except ImportError: 
    import dummy_threading as threading
    
try:
    from fastrlock.rlock import FastRLock as RLock
except ImportError:
    from threading import RLock

from pyxrd.generic.asynchronous.exceptions import *
from pyxrd.data.appdirs import user_log_dir

from . import settings
from .utils import start_script
from .status_thread import StatusThread

class Pyro4AsyncServerProvider(object):
    """
        Provider for the Pyro4 PyXRD server
    """
    
    _STATUS_NS_NOT_RUNNING = ("#FFFF00", "Nameserver Error", "Pyro4 Nameserver not running")
    _STATUS_ERR_NS_RUNNING = ("#FF0000", "Nameserver Exception", "Exception when checking if Pyro4 Nameserver is running")
    
    _STATUS_NS_NOT_LISTED = ("#FFFF00", "Nameserver Error", "Pyro4 PyXRD server not listed")
    _STATUS_ERR_NS_LISTED = ("#FF0000", "Nameserver Exception", "Exception when checking if Pyro4 PyXRD server is listed")
    
    _STATUS_NO_CONN_PYXRD_SERVER  = ("#FFFF00", "PyXRD Server Error", "Cannot connect to Pyro4 PyXRD server")
    _STATUS_ERR_CONN_PYXRD_SERVER = ("#FF0000", "PyXRD Server Exception", "Exception when connecting to Pyro4 PyXRD server")
    
    _STATUS_SUCCESS = ("#00FF00", "Connected (Pyro4)", "Succesfully connected to Pyro4 PyXRD Server")
    
    _updater = None
    
    status = _STATUS_NS_NOT_RUNNING
    status_lock = RLock()
    
    NS_CACHE_TIMEOUT = 30
    ns = None
    ns_ts = 0
    @classmethod
    def _locate_ns(cls):
        if cls.ns is None or time.time() - cls.ns_ts >= cls.NS_CACHE_TIMEOUT:
            cls.ns = Pyro4.naming.locateNS()
            cls.ns_ts = time.time()
        return cls.ns
    
    PROXY_CACHE_TIMEOUT = 30
    proxy = None
    proxy_ts = 0
    @classmethod
    def _get_proxy(cls):
        if cls.proxy is None or time.time() - cls.proxy_ts >= cls.PROXY_CACHE_TIMEOUT:
            if cls.proxy is None:
                cls.proxy = Pyro4.Proxy("PYRONAME:%s" % settings.PYRO_NAME)
            else:
                cls.proxy.pyroBind()
            cls.proxy_ts = time.time()
        return cls.proxy
    
    
    @classmethod
    def check_nameserver_alive(cls):
        try:
            ns = cls._locate_ns()
            ns.ping()
            return True
        except:
            print_exc()
            return False
    
    @classmethod
    def check_server_is_listed(cls):
        try:
            ns = cls._locate_ns()
            objects = ns.list()
            if settings.PYRO_NAME in objects:
                return True
            else:
                return False
        except:
            print_exc()
            return False
    
    """
        Async Provider Implementation: 
    """
    @classmethod
    def get_status(cls):
        """ 
            Should return a three-tuple consisting of the status colour, label and a description:
                ("#FF0000", "Error", "Nameserver not running")
            Status is updated periodically.
        """
        # first call:
        if cls._updater == None:
            cls._update_status(cls._get_status())
            cls._updater = StatusThread(5, cls)
            cls._updater.setDaemon(True)
            cls._updater.start()
        
        return cls.status
    
    @classmethod
    def _update_status(cls, status):
        with cls.status_lock:
            cls.status = status
    
    @classmethod
    def _get_status(cls):
        try:
            if not cls.check_nameserver_alive():
                return  cls._STATUS_NS_NOT_RUNNING
        except:
            print_exc()
            return cls._STATUS_ERR_NS_RUNNING
        try:         
            if not cls.check_server_is_listed():
                return cls._STATUS_NS_NOT_LISTED
        except:
            print_exc()
            return cls._STATUS_ERR_NS_LISTED
        try:         
            if not cls.check_server_is_alive():
                return cls._STATUS_NO_CONN_PYXRD_SERVER
        except:
            print_exc()
            return cls._STATUS_ERR_CONN_PYXRD_SERVER
        return cls._STATUS_SUCCESS
    
    @classmethod
    def check_server_is_alive(cls):
        try:
            server = cls.get_server(auto_start=False)
            return server.loopCondition()
        except:
            logging.info("Pyro4 PyXRD server not (yet) running!")
            return False
    
    @classmethod
    def get_server(cls, auto_start=True):
        if auto_start: 
            cls.launch_server()
        try:
            return Pyro4.Proxy("PYRONAME:%s" % settings.PYRO_NAME)
        except:
            print_exc()
            logging.error("Could not connect to Pyro4 PyXRD server.")
            raise ServerNotRunningException("Pyro4 PyXRD Server is not running!")
    
    @classmethod
    def launch_server(cls):
        if not cls.check_server_is_alive():
            log_file = os.path.join(user_log_dir('PyXRD'), 'server.log')
            start_script("run_server.py", auto_kill=not settings.KEEP_SERVER_ALIVE, log_file=log_file)
            ttl = 15
            delay = 0.2
            while not cls.check_server_is_alive():
                time.sleep(delay) # wait
                ttl -= 1
                if ttl == 0:
                    raise ServerStartTimeoutExcecption("Pyro4 PyXRD Server is not running!")
            logging.info("Pyro4 PyXRD server is running!")
            if not settings.KEEP_SERVER_ALIVE:
                atexit.register(cls.stop_server)
        
    @classmethod
    def stop_server(cls):
        try:
            server = cls.get_server(auto_start=False)
            server.shutdown()
        except:
            logging.error("Error when trying to shut down Pyro4 PyXRD server!")
            print_exc()
            raise ServerNotRunningException("Pyro4 PyXRD Server is not running!")
        
    pass #end of class



#!/usr/bin/env python3

import os
from setuptools import setup, find_packages

def get_version():
    from pyxrd.__version import __version__
    if __version__.startswith("v"):
        __version__ = __version__.replace("v", "")
    return "%s" % __version__

def get_install_requires():
    return [
        'setuptools',
        'numpy>=1.11',
        'scipy>=1.1.0',
        'matplotlib>=2.2.2',
        'Pyro4>=4.41',
        'deap>=1.0.1',
        'cairocffi',
        'pygobject>=3.20'
    ]

def read(fname):
    with open(os.path.join(os.path.dirname(__file__), fname)) as f:
        return f.read()

setup(
    name="PyXRD",
    version=get_version(),
    description="PyXRD is a python implementation of the matrix algorithm developed for the X-ray diffraction analysis of disordered lamellar structures",
    long_description=read('README.md'),
    keywords="XRD disorder mixed-layers",
    author="Mathijs Dumon",
    author_email="mathijs.dumon@gmail.com",
    url="http://github.org/mathijs-dumon/PyXRD",

    license="BSD",
    setup_requires=[ "setuptools_git >= 1.2", ],
    packages=find_packages(exclude=["test.*", "test", "tests_mvc", "tests_mvc.*"]),
    include_package_data=True,
    install_requires=get_install_requires(),
    zip_safe=False,

    classifiers=[
        "Development Status :: 4 - Beta",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3.4",
        "Environment :: Win32 (MS Windows)",
        "Environment :: X11 Applications :: Gnome",
        "Environment :: X11 Applications :: GTK",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: Science/Research",
        "Topic :: Utilities",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Visualization",
        "Natural Language :: English",
        "License :: OSI Approved :: BSD License",
    ],
)



# coding=UTF-8
# ex:ts=4:sw=4:et

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from mvc.support.utils import get_new_uuid


__all__ = [
    "gobject",
    "GtkTreeIter",
    "GenericTreeModel"
    "TREE_MODEL_LIST_ONLY"
]

TREE_MODEL_LIST_ONLY = 0x00
TREE_MODEL_ITERS_PERSIST = 0x00

events_pending = lambda: False

class GtkTreeIter():
    def __init__(self, user_data, path=None):
        self.user_data = user_data
        self.path = path

    pass # end of class




class GenericTreeModel(object):
    __connected_signals__ = None

    def __init__(self):
        self.__connected_signals__ = {}

    def connect(self, signal_name, handler, *args):
        handlers = self.__connected_signals__.get(signal_name, {})
        handler_id = get_new_uuid()
        handlers[handler_id] = (handler, args)
        self.__connected_signals__[signal_name] = handlers
        return handler_id

    def disconnect(self, signal_name, handler_id):
        try:
            handlers = self.__connected_signals__.get(signal_name, {})
            del handlers[handler_id]
        except KeyError:
            pass
        return

    def emit(self, signal_name, args=()):
        handlers = self.__connected_signals__.get(signal_name, {})
        for id, (handler, user_args) in handlers.items(): # @ReservedAssignment
            handler(self, *((args,) + user_args))
        pass

    def set_property(self, *args, **kwargs):
        pass

    def create_tree_iter(self, user_data):
        return GtkTreeIter(user_data)

    def get_path(self, itr):
        return self.on_get_path(itr.user_data)

    def get_iter(self, path):
        return self.create_tree_iter(self.on_get_iter(path))

    def row_inserted(self, path, itr):
        self.emit("row-inserted", (path, itr))

    def row_deleted(self, indeces):
        self.emit("row-deleted", (indeces,))

    def invalidate_iters(self):
        pass # TOD0!

    def iter_is_valid(self, itr):
        return True # TODO!

    def __len__(self):
        return len(self._model_data)

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import hashlib

from pyxrd.generic.models import PyXRDModel
from pyxrd.generic.io.json_codec import PyXRDEncoder

from mvc.models.properties.signal_property import SignalProperty
from mvc.models.properties.labeled_property import LabeledProperty

class AppModel(PyXRDModel):
    """
        Simple model that stores the state of the application window.
        Should never be made persistent.
        
        Attributes:
            needs_plot_update: a mvc.Signal to indicate the plot needs an
                update. This models listens for the 'needs_update' signal on the
                loaded project and propagates this accordingly.
            current_project: the currently loaded project
            current_specimen: the currently selected specimen, is None if more
                than one specimen is selected.
            current_specimens: a list of currently selected specimens, is never
                None, even if only one specimen is selected.
            single_specimen_selected: a boolean indicating whether or not a
                single specimen is selected
            multiple_specimen_selected: a boolean indicating whether or not
                multiple specimen are selected
    """

    # SIGNALS:
    needs_plot_update = SignalProperty()

    # PROPERTIES:
    current_project = LabeledProperty(default=None, text="Current project")
    
    @current_project.setter
    def current_project(self, value):
        _current_project = type(self).current_project._get(self)
        if _current_project is not None: 
            self.relieve_model(_current_project)
            self._project_hash = None
        _current_project = value
        type(self).current_project._set(self, value)
        type(type(self)).object_pool.clear()
        if _current_project is not None: self.observe_model(_current_project)
        self.clear_selected()
        self.needs_plot_update.emit()

    @property
    def current_filename(self):
        return self.current_project.filename if self.current_project else None

    current_specimen = LabeledProperty(default=None, text="Current specimen")
    @current_specimen.setter
    def current_specimen(self, value):
        type(self).current_specimens._set(self, [value])
        type(self).current_specimen._set(self, value)

    current_specimens = LabeledProperty(default=[], text="Current specimens")
    @current_specimens.setter
    def current_specimens(self, value):
        if value == None:
            value = []
        type(self).current_specimens._set(self, value)
        if len(self._current_specimens) == 1:
            type(self).current_specimen._set(self, self._current_specimens[0])
        else:
            type(self).current_specimen._set(self, None)

    @property
    def project_loaded(self):
        return self.current_project is not None

    @property
    def specimen_selected(self):
        return self.current_specimen is not None

    @property
    def single_specimen_selected(self):
        return self.specimen_selected and len(self.current_specimens) == 1

    @property
    def multiple_specimens_selected(self):
        return len(self.current_specimens) > 1

    # ------------------------------------------------------------
    #      Initialization and other internals
    # ------------------------------------------------------------
    def __init__(self, project=None):
        """ Initializes the AppModel with the given Project. """
        super(AppModel, self).__init__()
        self.current_project = project
        if project: project.parent = self

    # ------------------------------------------------------------
    #      Notifications of observable properties
    # ------------------------------------------------------------
    @PyXRDModel.observe("data_changed", signal=True)
    @PyXRDModel.observe("visuals_changed", signal=True)
    def notify_needs_update(self, model, prop_name, info):
        self.needs_plot_update.emit()

    def clear_selected(self):
        self.current_specimens = None

    # ------------------------------------------------------------
    #      Project change management:
    # ------------------------------------------------------------
    _project_hash = None
    def update_project_last_save_hash(self):
        if self.current_project is not None:
            dump = PyXRDEncoder.dump_object(self.current_project).encode(errors='ignore')
            self._project_hash = hashlib.sha224(dump).hexdigest()
    def check_for_changes(self):
        current_hash = None
        if self.current_project is not None:
            dump = PyXRDEncoder.dump_object(self.current_project).encode(errors='ignore')
            current_hash = hashlib.sha224(dump).hexdigest()
        return current_hash != self._project_hash

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from itertools import chain
from math import pi

import numpy as np

from .atoms import get_structure_factor

def calculate_z(default_z, lattice_d, z_factor):
    return lattice_d + z_factor * (default_z - lattice_d)

def get_factors(range_stl, component):
    r"""
    Returns a tuple containing the structure factor and phase difference for
        - range_stl: numpy array of 2*sin(θ) / λ values
        - component: a :class:`~pyxrd.calculations.data_objects.ComponentData`
          instance containing two lists of layer and interlayer
          :class:`~pyxrd.calculations.data_objects.Atom`'s.
          
    This function calls 
    :meth:`~pyxrd.calculations.atoms.get_structure_factor` for each
    atom in the layer and interlayer list of the component data object and
    sums the result to obtain the component's structure factors.
    
    The component's phase differences are calculated as follows:
      
    .. math::
        :nowrap:
        
        \begin{flalign*}
            & \psi = e^ { 2 \pi \cdot { \frac { 2 \cdot sin(\theta)} {\lambda} } \cdot \left( d_{001} \cdot i - \pi \cdot \delta d_{001} \cdot { \frac {2 \cdot sin(\theta)} {\lambda} } \right) } 
        \end{flalign*} 
        
    """

    z_factor = (component.d001 - component.lattice_d) / (component.default_c - component.lattice_d)

    num_layer_atoms = len(component.layer_atoms)

    sf_tot = 0.0 + 0.0j
    for i, atom in enumerate(chain(component.layer_atoms, component.interlayer_atoms)):
        atom.z = atom.default_z
        if i >= num_layer_atoms:
            atom.z = calculate_z(atom.z, component.lattice_d, z_factor)
        sf_tot += get_structure_factor(range_stl, atom)

    phi_tot = np.exp(2.*pi * range_stl * (component.d001 * 1j - pi * component.delta_c * range_stl))
    return sf_tot, phi_tot



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from pkg_resources import resource_filename # @UnresolvedImport

import gi
gi.require_version('Gtk', '3.0')  # @UndefinedVariable
from gi.repository import Gtk  # @UnresolvedImport

from matplotlib.figure import Figure
from matplotlib.backends.backend_gtk3cairo import FigureCanvasGTK3Cairo as FigureCanvasGTK

from pyxrd.generic.views import BaseView, HasChildView, DialogView,\
    ObjectListStoreView
from mvc.adapters.gtk_support.widgets import ScaleEntry

class EditPhaseView(HasChildView, BaseView):
    title = "Edit Phases"
    builder = resource_filename(__name__, "glade/phase.glade")
    top = "edit_phase"
    widget_format = "phase_%s"

    csds_view = None
    csds_view_container = widget_format % "CSDS_distribution"

    probabilities_view = None
    probabilities_view_container = widget_format % "probabilities"

    components_view = None
    components_view_container = widget_format % "components"

    def set_csds_view(self, view):
        self.csds_view = view
        if view is not None:
            self._add_child_view(view.get_top_widget(), self[self.csds_view_container])
        return view

    def set_csds_sensitive(self, sens):
        self[self.csds_view_container].set_sensitive(sens)

    def set_probabilities_view(self, view):
        self.probabilities_view = view
        if view is not None:
            self._add_child_view(view.get_top_widget(), self[self.probabilities_view_container])
        return view

    def remove_probabilities(self):
        num = self["book_wrapper"].page_num(self[self.probabilities_view_container])
        self["book_wrapper"].remove_page(num)

    def set_components_view(self, view):
        self.components_view = view
        if view is not None:
            self._add_child_view(view.get_top_widget(), self[self.components_view_container])
        return view

class EditRawPatternPhaseView(BaseView):
    title = "Edit Raw Pattern Phase"
    builder = resource_filename(__name__, "glade/raw_pattern_phase.glade")
    top = "edit_raw_pattern_phase"
    widget_format = "rp_phase_%s"

class EditAtomRatioView(DialogView):
    title = "Edit Atom Ratio"
    subview_builder = resource_filename(__name__, "glade/ratio.glade")
    subview_toplevel = "edit_ratio"
    modal = True
    widget_format = "ratio_%s"

    @property
    def atom1_combo(self):
        return self["ratio_atom1"]

    @property
    def atom2_combo(self):
        return self["ratio_atom2"]

    pass # end of class

class EditAtomContentsView(DialogView):
    title = "Edit Atom Contents"
    subview_builder = resource_filename(__name__, "glade/contents.glade")
    subview_toplevel = "edit_contents"
    modal = True
    widget_format = "contents_%s"

    contents_list_view_container = "atom_contents_container"

    def set_contents_list_view(self, view):
        self[self.widget_format % "atom_contents"] = view
        return self._add_child_view(view, self[self.contents_list_view_container])

    @property
    def atom_contents_container(self):
        return self["container_atom_contents"]

    pass # end of class

class EditComponentView(HasChildView, BaseView):
    title = "Edit Component"
    builder = resource_filename(__name__, "glade/component.glade")
    top = "edit_component"
    widget_format = "component_%s"

    layer_view = None
    layer_view_container = widget_format % "layer_atoms"

    interlayer_view = None
    interlayer_view_container = widget_format % "interlayer_atoms"

    atom_relations_view = None
    atom_relations_view_container = widget_format % "atom_relations"

    ucpa_view = None
    ucpa_view_container = widget_format % "ucp_a"

    ucpb_view = None
    ucpb_view_container = widget_format % "ucp_b"


    def __init__(self, *args, **kwargs):
        BaseView.__init__(self, *args, **kwargs)

    def set_layer_view(self, view):
        self.layer_view = view
        return self._add_child_view(view, self[self.layer_view_container])

    def set_atom_relations_view(self, view):
        self.atom_relations_view = view
        return self._add_child_view(view, self[self.atom_relations_view_container])

    def set_interlayer_view(self, view):
        self.interlayer_view = view
        return self._add_child_view(view, self[self.interlayer_view_container])

    def set_ucpa_view(self, view):
        self.ucpa_view = view
        return self._add_child_view(view, self[self.ucpa_view_container])

    def set_ucpb_view(self, view):
        self.ucpb_view = view
        return self._add_child_view(view, self[self.ucpb_view_container])

class EditUnitCellPropertyView(BaseView):
    builder = resource_filename(__name__, "glade/unit_cell_prop.glade")
    top = "box_ucf"
    widget_format = "ucp_%s"

class EditCSDSDistributionView(BaseView):
    builder = resource_filename(__name__, "glade/csds.glade")
    top = "tbl_csds_distr"

    def __init__(self, *args, **kwargs):
        BaseView.__init__(self, *args, **kwargs)

        self.graph_parent = self["distr_plot_box"]
        self.setup_matplotlib_widget()

    def setup_matplotlib_widget(self):
        #style = Gtk.Style()
        self.figure = Figure(dpi=72) #, edgecolor=str(style.bg[2]), facecolor=str(style.bg[2]))

        self.plot = self.figure.add_subplot(111)
        self.figure.subplots_adjust(bottom=0.20)

        self.matlib_canvas = FigureCanvasGTK(self.figure)

        self.plot.autoscale_view()

        self.graph_parent.add(self.matlib_canvas)
        self.graph_parent.show_all()

    def update_figure(self, distr):
        self.plot.cla()
        self.plot.hist(list(range(len(distr))), len(distr), weights=distr, normed=1, ec='b', histtype='stepfilled')
        self.plot.set_ylabel('')
        self.plot.set_xlabel('CSDS', size=14, weight="heavy")
        self.plot.relim()
        self.plot.autoscale_view()
        if self.matlib_canvas is not None:
            self.matlib_canvas.draw()

    def reset_params(self):
        tbl = self["tbl_params"]
        for child in tbl.get_children():
            tbl.remove(child)
        tbl.resize(1, 2)

    def add_param_widget(self, name, label, minimum, maximum):
        tbl = self["tbl_params"]
        rows = tbl.get_property("n-rows") + 1
        tbl.resize(rows, 2)

        lbl = Gtk.Label(label=label)
        lbl.set_alignment(1.0, 0.5)
        tbl.attach(lbl, 0, 1, rows - 1, rows, Gtk.AttachOptions.FILL, Gtk.AttachOptions.FILL)

        inp = ScaleEntry(minimum, maximum, enforce_range=True)
        tbl.attach(inp, 1, 2, rows - 1, rows, Gtk.AttachOptions.FILL, Gtk.AttachOptions.FILL)

        tbl.show_all()

        self[name] = inp
        inp.set_name(name)

        return inp

class AddPhaseView(DialogView):
    title = "Add Phase"
    subview_builder = resource_filename(__name__, "glade/addphase.glade")
    subview_toplevel = "add_phase_container"

    active_type = "empty" # | default | raw

    def __init__(self, *args, **kwargs):
        DialogView.__init__(self, *args, **kwargs)

    def get_G(self):
        return int(self["G"].get_value_as_int())

    def get_R(self):
        return int(self["R"].get_value_as_int())

    def get_phase_type(self):
        if self.active_type == "empty":
            return "empty"
        elif self.active_type == "default":
            itr = self.phase_combo_box.get_active_iter()
            val = self.phase_combo_box.get_model().get_value(itr, 1) if itr else None
            return val
        else:
            return "raw"

    def update_sensitivities(self):
        self["cont_empty_phase"].set_sensitive(False)
        self["cont_default_phase"].set_sensitive(False)
        if self["rdb_empty_phase"].get_active():
            self.active_type = "empty"
            self["cont_empty_phase"].set_sensitive(True)
        elif self["rdb_default_phase"].get_active():
            self.active_type = "default"
            self["cont_default_phase"].set_sensitive(True)
        else:
            self.active_type = "raw"

    @property
    def phase_combo_box(self):
        return self["cmb_default_phases"]

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from math import exp, sqrt, log, pi

import numpy as np

sqrtpi = sqrt(pi)
sqrt2pi = sqrt(2 * pi)
sqrt8 = sqrt(8)

def mmult(A, B):
    return np.sum(np.transpose(A, (0, 2, 1))[:, :, :, np.newaxis] * B[:, :, np.newaxis, :], -3)

def mdot(A, B):
    C = np.zeros(shape=A.shape, dtype=np.complex)
    for i in range(A.shape[0]):
        C[i] = np.dot(A[i], B[i])
    return C

def mtim(A, B): # currently unused?
    C = np.zeros(shape=A.shape, dtype=np.complex)
    for i in range(A.shape[0]):
        C[i] = np.multiply(A[i], B[i])
    return C

def solve_division(A, B): # currently unused?
    bt = np.transpose(B, axes=(0, 2, 1))
    at = np.transpose(A, axes=(0, 2, 1))
    return np.array([np.transpose(np.linalg.lstsq(bt[i], at[i])[0]) for i in range(bt.shape[0])])

def lognormal(T, a, b):
    return exp(-(log(T) - a) ** 2 / (2.0 * (b ** 2))) / (sqrt2pi * abs(b) * T)

def add_noise(x, noise_fraction=0.05):
    if x.size > 0:
        abs_value = noise_fraction * np.amax(x)
        return x + np.random.standard_normal(x.shape) * abs_value
    else:
        return x

def smooth(x, half_window_len=3, window='blackman'):
    """smooth the data using a window with requested size.
    
    This method is based on the convolution of a scaled window with the signal.
    The signal is prepared by introducing reflected copies of the signal 
    (with the window size) in both ends so that transient parts are minimized
    in the beginning and end part of the output signal.
    
    input:
        x: 1D array like (assumed to be spaced equally)
        half_window_len: half of the dimension of the smoothing window, actual window
            is calculated from this so that: window_len = 2*half_window_len + 1, this
            ensures that the window is always an odd number, regardless of user input;
        window: the type of window from 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'
            flat window will produce a moving average smoothing.

    output:
        the smoothed signal
        
    example:

    t=linspace(-2,2,0.1)
    x=sin(t)+randn(len(t))*0.1
    y=smooth(x)
    
    see also: 
    
    numpy.hanning, numpy.hamming, numpy.bartlett, numpy.blackman, numpy.convolve
    scipy.signal.lfilter
 
    """

    window_len = half_window_len * 2 + 1

    if x.ndim != 1:
        x = np.ndarray.flatten(x)

    if x.size < window_len:
        raise ValueError("Input vector needs to be bigger than window size.")

    if window_len < 3:
        return x

    if not window in ['flat', 'hanning', 'hamming', 'bartlett', 'blackman']:
        raise ValueError("Window is on of 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'")


    s = np.r_[x[window_len - 1:0:-1], x, x[-1:-window_len:-1]]
    if window == 'flat': # moving average
        w = np.ones(window_len, 'd')
    else:
        w = eval('np.' + window + '(window_len)')

    y = np.convolve(w / w.sum(), s, mode='valid')
    return y[half_window_len:-half_window_len]


# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import os, locale
import logging
logger = logging.getLogger(__name__)

from mvc.adapters.gtk_support.dialogs.dialog_factory import DialogFactory
from mvc.adapters.gtk_support.tree_view_adapters import wrap_xydata_to_treemodel, \
    wrap_list_property_to_treemodel

from pyxrd.generic.controllers import BaseController
from pyxrd.generic.controllers.objectliststore_controllers import TreeViewMixin
from pyxrd.generic.views.treeview_tools import setup_treeview, new_text_column

from ..models import RawPatternPhase

class EditRawPatternPhaseController(TreeViewMixin, BaseController):
    """ 
        Controller for the phase edit view
    """

    file_filters = RawPatternPhase.Meta.rp_filters
    rp_export_filters = RawPatternPhase.Meta.rp_export_filters

    widget_handlers = {
        'custom': 'custom_handler',
    }

    @property
    def phases_treemodel(self):
        if self.model.project is not None:
            prop = self.model.project.Meta.get_prop_intel_by_name("phases")
            return wrap_list_property_to_treemodel(self.model.project, prop)
        else:
            return None

    # ------------------------------------------------------------
    #      Initialization and other internals
    # ------------------------------------------------------------
    @staticmethod
    def custom_handler(self, intel, widget):
        pass # nothing to do

    def setup_raw_pattern_tree_view(self, store, widget):
        """
            Creates the raw pattern TreeView layout and behavior
        """

        setup_treeview(widget, store,
            on_cursor_changed=self.on_raw_pattern_tv_cursor_changed,
            sel_mode='MULTIPLE')
        # X Column:
        widget.append_column(new_text_column(
            '°2θ', text_col=store.c_x, editable=True,
            edited_callback=(self.on_xy_data_cell_edited, (self.model.raw_pattern, 0)),
            resizable=True, expand=True))
        # Y Column:
        widget.append_column(new_text_column(
            'Intensity', text_col=store.c_y, editable=True,
            edited_callback=(self.on_xy_data_cell_edited, (self.model.raw_pattern, 1)),
            resizable=True, expand=True))

    # ------------------------------------------------------------
    #      Methods & Functions
    # ------------------------------------------------------------
    def get_raw_pattern_tree_model(self):
        return wrap_xydata_to_treemodel(self.model, self.model.Meta.get_prop_intel_by_name("raw_pattern"))

    # ------------------------------------------------------------
    #      Notifications of observable properties
    # ------------------------------------------------------------
    @BaseController.observe("name", assign=True)
    def notif_name_changed(self, model, prop_name, info):
        self.phases_treemodel.on_item_changed(self.model)
        return

    # ------------------------------------------------------------
    #      GTK Signal handlers
    # ------------------------------------------------------------
    def on_raw_pattern_tv_cursor_changed(self, tv):
        path, _ = tv.get_cursor()
        self.view["btn_del_raw_data"].set_sensitive(path is not None)
        return True

    def on_add_raw_pattern_clicked(self, widget):
        self.model.raw_pattern.append(0, 0)
        return True

    def on_del_raw_pattern_clicked(self, widget):
        paths = self.get_selected_paths(self.view["phase_rp_raw_pattern"])
        if paths is not None:
            self.model.raw_pattern.remove_from_indeces(*paths)
        return True

    def on_xy_data_cell_edited(self, cell, path, new_text, model, col):
        try:
            value = float(locale.atof(new_text))
        except ValueError:
            logger.exception("ValueError: Invalid literal for float(): '%s'" % new_text)
        else:
            model.set_value(int(path), col, value)
        return True

    def on_replace_raw_pattern(self, *args, **kwargs):
        def on_accept(dialog):
            filename = dialog.filename
            parser = dialog.parser
            try:
                self.model.raw_pattern.load_data(parser, filename, clear=True)
            except Exception:
                message = "An unexpected error has occured when trying to parse '%s'.\n" % os.path.basename(filename)
                message += "This is most likely caused by an invalid or unsupported file format."
                DialogFactory.get_information_dialog(
                    message=message, parent=self.view.get_toplevel()
                ).run()
                raise
        DialogFactory.get_load_dialog(
            "Open XRD file for import", parent=self.view.get_toplevel(),
            filters=self.file_filters
        ).run(on_accept)
        return True

    def on_btn_import_raw_pattern_clicked(self, widget, data=None):
        def on_confirm(dialog):
            self.on_replace_raw_pattern()
        DialogFactory.get_confirmation_dialog(
            "Importing a new experimental file will erase all current data.\nAre you sure you want to continue?",
            parent=self.view.get_toplevel()
        ).run(on_confirm)
        return True

    def on_export_raw_pattern(self, *args, **kwargs):
        return self._export_data(self.model.raw_pattern)

    def on_btn_export_raw_pattern_clicked(self, widget, data=None):
        return self.on_export_raw_pattern()

    def _export_data(self, line):
        def on_accept(dialog):
            filename = dialog.filename
            parser = dialog.parser
            try:
                line.save_data(parser, filename, **self.model.get_export_meta_data())
            except Exception:
                message = "An unexpected error has occured when trying to save to '%s'." % os.path.basename(filename)
                DialogFactory.get_information_dialog(
                    message=message, parent=self.view.get_toplevel()
                ).run()
                raise
        ext_less_fname = os.path.splitext(self.model.name)[0]
        DialogFactory.get_save_dialog(
            "Select file for export", parent=self.view.get_toplevel(),
            filters=self.rp_export_filters,
            current_name=ext_less_fname
        ).run(on_accept)

    pass #end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

try:
    from lxml import ET
except ImportError:
    try:
        # Python 2.5
        import xml.etree.cElementTree as ET
    except ImportError:
        try:
            # Python 2.5
            import xml.etree.ElementTree as ET
        except ImportError:
            try:
                # normal cElementTree install
                import cElementTree as ET
            except ImportError:
                try:
                    # normal ElementTree install
                    import elementtree.ElementTree as ET
                except ImportError:
                    print("Failed to import ElementTree from any known place") 

class XMLParserMixin(object):
    """
        XML Parser Mixin class
    """

    @classmethod
    def get_xml_for_string(cls, s):
        """ Returns a tuple containing the XML tree and root objects """
        root = ET.fromstring(s)
        return ET.ElementTree(element=root), root

    @classmethod
    def get_xml_for_file(cls, f):
        """ Returns a tuple containing the XML tree and root objects """
        tree = ET.parse(f)
        return tree, tree.getroot()

    @classmethod
    def get_val(cls, root, path, attrib, default=None):
        """ Returns the attribute `attrib` from the first element found in
        `root` using the given `path`or default if not found """
        element = root.find(path)
        if element is not None:
            return element.get(attrib)
        else:
            return default

    pass #end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from .meta_parser import MetaParser
from .data_object import DataObject
import types

class BaseParser(object, metaclass=MetaParser):
    """
        Base class providing some common attributes and functions.
        Do not register this class or subclasses without overriding the
        following functions:
            - parse_header
            - parse_data
            - parse (optional)
            - setup_file_filter (optional)
    """

    # This should be changed by sub-classes
    description = "Base Parser"
    extensions = []
    mimetypes = []
    @property
    def can_write(self):
        return getattr(self, "write", None) is not None
    @property
    def can_read(self):
        return getattr(self, "parse", None) is not None
    data_object_type = DataObject

    # This should be changed by sub-classes
    __file_mode__ = "r"

    file_filter = None

    @classmethod
    def _get_file(cls, fp, close=None):
        """
            Returns a three-tuple:
            filename, file-object, close
        """
        if isinstance(fp, str):
            return fp, open(fp, cls.__file_mode__), True if close is None else close
        else:
            return getattr(fp, 'name', None), fp, False if close is None else close

    @classmethod
    def _adapt_data_object_list(cls, data_objects, num_samples, only_extend=False):
        # If not yet created, create data_objects list:
        if data_objects == None:
            data_objects = [None, ]
        # If not yet the same length, adapt:
        num_data_objects = len(data_objects)
        if num_data_objects < num_samples:
            data_objects.extend([None] * int(num_samples - num_data_objects))
        if not only_extend and num_data_objects > num_samples:
            data_objects = data_objects[:num_samples]
        # If not yet initialized, initialize:
        for i in range(num_samples):
            if not data_objects[i]:
                data_objects[i] = cls.data_object_type()
        return data_objects

    @classmethod
    def _parse_header(cls, filename, fp, data_objects=None, close=False):
        """
            This method is implemented by sub-classes.
            It should parse the file and returns a list of DataObjects 
            with the header properties filled in accordingly.
            The filename argument is always required. If no file object is passed
            as keyword argument, it only serves as a label. Otherwise a new file
            object is created. 
            File objects are not closed unless close is set to True.
            Existing DataObjects can be passed as well and will then 
            be used instead of creating new ones.
        """
        # This should be implemented by sub-classes
        raise NotImplementedError

    @classmethod
    def _parse_data(cls, filename, fp, data_objects=None, close=False):
        """
            This method is implemented by sub-classes.
            It should parse the file and return a list of DataObjects
            with the data properties filled in accordingly.
            The filename argument is always required. If no file object is passed
            as keyword argument, it only serves as a label. Otherwise a new file
            object is created.
            File objects are not closed unless close is set to True.
            Existing DataObjects can be passed as well and will then 
            be used instead of creating new ones.
        """
        # This should be implemented by sub-classes
        raise NotImplementedError


    @classmethod
    def parse(cls, fp, data_objects=None, close=True):
        """
            This method parses the file and return a list of DataObjects
            with both header and data properties filled in accordingly.
            The filename argument is always required. If no file object is passed
            as keyword argument, it only serves as a label. Otherwise a new file
            object is created.
            File objects are closed unless close is set to False.
            Existing DataObjects can be passed as well and will then 
            be used instead of creating new ones.
        """
        filename, fp, close = cls._get_file(fp, close=close)
        data_objects = cls._parse_header(filename, fp, data_objects=data_objects)
        data_objects = cls._parse_data(filename, fp, data_objects=data_objects)
        if close: fp.close()
        return data_objects

    @classmethod
    def setup_file_filter(cls):
        """
            Creates a file filter based on a list of extensions set in the
            'extensions' attribute of the class using the 'description' attribute
            as the name for the filter. If the 'mimetypes' attribute is also set,
            it will also set these. If additional properties are needed, this function
            should be overriden by subclasses.
        """
        if cls.file_filter == None and cls.description != "" and cls.extensions:
            try:
                import gi
                gi.require_version('Gtk', '3.0')
                from gi.repository import Gtk
            except ImportError:
                pass
            else:
                # Init file filter:
                cls.file_filter = Gtk.FileFilter()
                cls.file_filter.set_name(cls.description)
                for mtpe in cls.mimetypes:
                    # cls.file_filter.add_mime_type(mtpe)
                    pass
                for expr in cls.extensions:
                    cls.file_filter.add_pattern(expr)
                setattr(cls.file_filter, "parser", cls)

    pass # end of class


#!/usr/bin/python
# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import os
import unittest
from io import StringIO

from pyxrd.generic.io import storables, Storable

from pyxrd.file_parsers.json_parser import JSONParser

__all__ = [
    'TestParserMixin',
]


def load_data_from_files(*files):
    basepath = os.path.realpath(os.getcwd())
    for fname in files:
        with open(basepath + "/" + fname, 'rb') as f:
            yield f.read()

class TestStorable(unittest.TestCase):

    @storables.register()
    class DummyStorable(Storable):

        __storables__ = [
            "name",
            "data",
            "my_daddy"
        ]

        class Meta(Storable.Meta):
            store_id = "DummyStorable"

        def __init__(self, name, data, my_daddy):
            super(TestStorable.DummyStorable, self).__init__()
            self.name = name
            self.data = data
            self.my_daddy = my_daddy

    def setUp(self):
        self.daddy = self.DummyStorable('Daddy Dummy', list(range(50)), None)
        self.child = self.DummyStorable('Child Dummy', list(range(5)), self.daddy)

    def tearDown(self):
        del self.daddy
        del self.child

    def test_setup(self):
        self.assertNotEqual(self.daddy, None)
        self.assertNotEqual(self.child, None)

    def test_encoding(self):
        self.daddy_dump = self.daddy.dump_object()
        self.child_dump = self.child.dump_object()
        self.assertIn("".join(self.daddy_dump.split()), "".join(self.child_dump.split()))


    def test_decoding(self):

        child_encoded = """
{
    "type": "DummyStorable", 
    "properties": {
        "name": "Child Dummy", 
        "data": [
            0, 
            1, 
            2, 
            3, 
            4
        ], 
        "my_daddy": {
            "type": "DummyStorable", 
            "properties": {
                "name": "Daddy Dummy", 
                "data": [ 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10 ], 
                "my_daddy": null
            }
        }
    }
}"""
        f = StringIO(child_encoded)
        decoded_child = JSONParser.parse(f)

        self.assertNotEqual(decoded_child, None)
        self.assertNotEqual(decoded_child.my_daddy, None)

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from pkg_resources import resource_filename # @UnresolvedImport

import gi
from pyxrd.generic.plot.click_catcher import ClickCatcher
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk

import logging
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

import matplotlib
import matplotlib.transforms as transforms
from matplotlib.figure import Figure
from matplotlib.tight_layout import get_renderer
from matplotlib.backends.backend_gtk3 import NavigationToolbar2GTK3 as NavigationToolbar
from matplotlib.backends.backend_gtk3agg import FigureCanvasGTK3Agg as FigureCanvasGTK
try:
    from matplotlib.pyparsing import ParseFatalException
except ImportError:
    from pyparsing import ParseFatalException

from mpl_toolkits.axisartist import Subplot

from mvc.adapters.gtk_support.dialogs.dialog_factory import DialogFactory

from pyxrd.data import settings
from pyxrd.generic.plot.motion_tracker import MotionTracker
from pyxrd.generic.plot.axes_setup import PositionSetup, update_axes
from pyxrd.generic.plot.plotters import plot_specimens, plot_mixtures

class MainPlotController(object):
    """
        A controller for the main plot canvas.
        Sets up the widgets and has image exporting functionality.
    """
    
    file_filters = ("Portable Network Graphics (PNG)", "*.png"), \
                   ("Scalable Vector Graphics (SVG)", "*.svg"), \
                   ("Portable Document Format (PDF)", "*.pdf")

    _canvas = None
    @property
    def canvas(self):
        if not self._canvas:
            self.setup_figure()
            self.setup_canvas()
            self.setup_content()
        return self._canvas

    # ------------------------------------------------------------
    #      View integration getters
    # ------------------------------------------------------------
    def get_toolbar_widget(self, window):
        return NavigationToolbar(self.canvas, window)

    def get_canvas_widget(self):
        return self.canvas
    
    # ------------------------------------------------------------
    #      Initialization and other internals
    # ------------------------------------------------------------
    def __init__(self, status_callback, marker_callback, *args, **kwargs):
        self.setup_layout_cache()
        self.setup_figure()
        self.setup_canvas()
        self.setup_content(status_callback, marker_callback)

    def setup_layout_cache(self):
        self.position_setup = PositionSetup()
        self.labels = list()
        self.marker_lbls = list()
        self._proxies = dict()
        self.scale = 1.0
        self.stats = False
        self._last_pos = None

    def setup_figure(self):
        self.figure = Figure(dpi=72, facecolor="#FFFFFF", linewidth=0)
        self.figure.subplots_adjust(hspace=0.0, wspace=0.0)

    def setup_canvas(self):
        self._canvas = FigureCanvasGTK(self.figure)

    def setup_content(self, status_callback, marker_callback):
        # Create subplot and add it to the figure:
        self.plot = Subplot(self.figure, 211, facecolor=(1.0, 1.0, 1.0, 0.0))
        self.plot.set_autoscale_on(False)
        self.figure.add_axes(self.plot)

        # Connect events:
        self.canvas.mpl_connect('draw_event', self.fix_after_drawing)
        self.canvas.mpl_connect('resize_event', self.fix_after_drawing)

        self.mtc = MotionTracker(self, status_callback)
        self.cc = ClickCatcher(self, marker_callback)

        #self.update()

    # ------------------------------------------------------------
    #      Update methods
    # ------------------------------------------------------------
    def draw(self):
        self._last_pos = self.fix_before_drawing()
        self.figure.canvas.draw()

    def fix_after_drawing(self, *args):
        _new_pos = self.fix_before_drawing()
        
        if _new_pos != self._last_pos:
            self._last_pos = _new_pos
            self._redraw_later()

        return False

    def _redraw_later(self):
        self.timer = self.figure.canvas.new_timer(interval=10)
        self.timer.single_shot = True
        self.timer.add_callback(lambda : self.figure.canvas.draw_idle())
        self.timer.start()

    def fix_before_drawing(self, *args):
        """
            Fixes alignment issues due to longer labels or smaller windows
            Is executed after an initial draw event, since we can then retrieve
            the actual label dimensions and shift/resize the plot accordingly.
        """
        renderer = get_renderer(self.figure)        
        if not renderer or not self._canvas.get_realized():
            return False
        
        # Fix left side for wide specimen labels:
        if len(self.labels) > 0:
            bbox = self._get_joint_bbox(self.labels, renderer)
            if bbox is not None: 
                self.position_setup.left = self.position_setup.default_left + bbox.width
        # Fix top for high marker labels:
        if len(self.marker_lbls) > 0:
            bbox = self._get_joint_bbox([ label for label, flag, _ in self.marker_lbls if flag ], renderer)
            if bbox is not None: 
                self.position_setup.top = self.position_setup.default_top - bbox.height
        # Fix bottom for x-axis label:
        bottom_label = self.plot.axis["bottom"].label
        if bottom_label is not None:
            bbox = self._get_joint_bbox([bottom_label], renderer)
            if bbox is not None:
                self.position_setup.bottom = self.position_setup.default_bottom + (bbox.ymax - bbox.ymin) * 2.0 # somehow we need this?

        # Calculate new plot position & set it:
        plot_pos = self.position_setup.position
        self.plot.set_position(plot_pos)

        # Adjust specimen label position
        for label in self.labels:
            label.set_x(plot_pos[0] - 0.025)

        # Adjust marker label position
        for label, flag, y_offset in self.marker_lbls:
            if flag:
                newy = plot_pos[1] + plot_pos[3] + y_offset - 0.025
                label.set_y(newy)
        
        _new_pos = self.position_setup.to_string()
        return _new_pos
    
    def update(self, clear=False, project=None, specimens=None):
        """
            Updates the entire plot with the given information.
        """
        if clear: self.plot.cla()

        if project and specimens:
            self.labels, self.marker_lbls = plot_specimens(
                self.plot, self.position_setup, self.cc,
                project, specimens
            )
            # get mixtures for the selected specimens:
            plot_mixtures(self.plot, project, [ mixture for mixture in project.mixtures if any(specimen in mixture.specimens for specimen in specimens) ])

        update_axes(
            self.plot, self.position_setup,
            project, specimens
        )

        self.draw()

    # ------------------------------------------------------------
    #      Plot position and size calculations
    # ------------------------------------------------------------
    def _get_joint_bbox(self, container, renderer):
        bboxes = []
        try:
            for text in container:
                bbox = text.get_window_extent(renderer=renderer)
                # the figure transform goes from relative coords->pixels and we
                # want the inverse of that
                bboxi = bbox.inverse_transformed(self.figure.transFigure)
                bboxes.append(bboxi)
        except (RuntimeError, ValueError):
            logger.exception("Caught unhandled exception when joining boundig boxes")
            return None # don't continue
        # this is the bbox that bounds all the bboxes, again in relative
        # figure coords
        if len(bboxes) > 0:
            bbox = transforms.Bbox.union(bboxes)
            return bbox
        else:
            return None

    # ------------------------------------------------------------
    #      Graph exporting
    # ------------------------------------------------------------
    def save(self, parent=None, current_name="graph", size="auto", num_specimens=1, offset=0.75):
        """
            Displays a save dialog to export an image from the current plot.
        """
        # Parse arguments:
        width, height = 0, 0
        if size == "auto":
            descr, width, height, dpi = settings.OUTPUT_PRESETS[0]
        else:
            width, height, dpi = list(map(float, size.replace("@", "x").split("x")))

        # Load gui:
        builder = Gtk.Builder()
        builder.add_from_file(resource_filename("pyxrd.specimen", "glade/save_graph_size.glade")) # FIXME move this to this namespace!!
        size_expander = builder.get_object("size_expander")
        cmb_presets = builder.get_object("cmb_presets")

        # Setup combo with presets:
        cmb_store = Gtk.ListStore(str, int, int, float)
        for row in settings.OUTPUT_PRESETS:
            cmb_store.append(row)
        cmb_presets.clear()
        cmb_presets.set_model(cmb_store)
        cell = Gtk.CellRendererText()
        cmb_presets.pack_start(cell, True)
        cmb_presets.add_attribute(cell, 'text', 0)
        def on_cmb_changed(cmb, *args):
            itr = cmb.get_active_iter()
            w, h, d = cmb_store.get(itr, 1, 2, 3)
            entry_w.set_text(str(w))
            entry_h.set_text(str(h))
            entry_dpi.set_text(str(d))
        cmb_presets.connect('changed', on_cmb_changed)

        # Setup input boxes:
        entry_w = builder.get_object("entry_width")
        entry_h = builder.get_object("entry_height")
        entry_dpi = builder.get_object("entry_dpi")
        entry_w.set_text(str(width))
        entry_h.set_text(str(height))
        entry_dpi.set_text(str(dpi))

        # What to do when the user wants to save this:
        def on_accept(dialog):
            # Get the width, height & dpi
            width = float(entry_w.get_text())
            height = float(entry_h.get_text())
            dpi = float(entry_dpi.get_text())
            i_width, i_height = width / dpi, height / dpi
            # Save it all right!
            self.save_figure(dialog.filename, dpi, i_width, i_height)

        # Ask the user where, how and if he wants to save:
        DialogFactory.get_save_dialog(
            "Save Graph", parent=parent,
            filters=self.file_filters, current_name=current_name,
            extra_widget=size_expander
        ).run(on_accept)

    def save_figure(self, filename, dpi, i_width, i_height):
        """
            Save the current plot
            
            Arguments:
             filename: the filename to save to (either .png, .pdf or .svg)
             dpi: Dots-Per-Inch resolution
             i_width: the width in inch
             i_height: the height in inch
        """
        # Get original settings:
        original_dpi = self.figure.get_dpi()
        original_width, original_height = self.figure.get_size_inches()
        # Set everything according to the user selection:
        self.figure.set_dpi(dpi)
        self.figure.set_size_inches((i_width, i_height))
        self.figure.canvas.draw() # replot
        bbox_inches = matplotlib.transforms.Bbox.from_bounds(0, 0, i_width, i_height)
        # Save the figure:
        self.figure.savefig(filename, dpi=dpi, bbox_inches=bbox_inches)
        # Put everything back the way it was:
        self.figure.set_dpi(original_dpi)
        self.figure.set_size_inches((original_width, original_height))
        self.figure.canvas.draw() # replot

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from pkg_resources import resource_filename # @UnresolvedImport

import gi
gi.require_version('Gtk', '3.0')  # @UndefinedVariable
from gi.repository import Gtk, GdkPixbuf  # @UnresolvedImport

from pyxrd.data import settings

from pyxrd.generic.views import ObjectListStoreView, BaseView, HasChildView, FormattedTitleView

from pyxrd.project.views import ProjectView
from pyxrd.specimen.views import SpecimenView, EditMarkersView
from pyxrd.application.icons import get_icon_list

class AppView(HasChildView, FormattedTitleView):
    """
        The main application interface view.
        
        Attributes:
            project: the project view
            specimen: the specimen view
            markers: the markers view
            phases: the phases view
            atom_types: the atom_types view
            statistics: the statistics view
            mixtures: the mixtures view
            
        
    """
    builder = resource_filename(__name__, "glade/application.glade")
    top = "main_window"
    title_format = "PyXRD - %s"

    child_views = {
        "project": ProjectView,
        "specimen": SpecimenView,
        "markers": EditMarkersView, # FIXME this should be part of the specimen view/controller code
        "phases": ObjectListStoreView,
        "atom_types": ObjectListStoreView,
        "behaviours": ObjectListStoreView,
        "mixtures": ObjectListStoreView
    }

    widget_groups = {
        'full_mode_only': [
            "tbtn_edit_phases",
            #"tbtn_edit_behaviours",
            "tbtn_edit_atom_types",
            "tbtn_edit_mixtures",
            "tbtn_separator1",
            "btn_sample",
            "separator3",
            "separator4",
            "separator5",
            "main_menu_item_edit_phases",
            "main_menu_item_edit_atom_types",
            "main_menu_item_edit_mixtures"
        ]
    }

    # ------------------------------------------------------------
    #      Initialisation and other internals
    # ------------------------------------------------------------
    def __init__(self, *args, **kwargs):
        super(AppView, self).__init__(*args, **kwargs)

        # Setup about window:
        def on_aboutbox_response(dialog, response, *args):
            if response < 0:
                dialog.hide()
                dialog.emit_stop_by_name('response')

        def on_aboutbox_close(widget, event=None):
            self["about_window"].hide()
            return True

        self["about_window"].set_version(settings.VERSION)
        self["about_window"].set_website("https://github.com/PyXRD/PyXRD/blob/v%s/Manual.pdf" % settings.VERSION);
        pixbuf = GdkPixbuf.Pixbuf.new_from_file(resource_filename(__name__, "icons/pyxrd.png")) # @UndefinedVariable
        scaled_buf = pixbuf.scale_simple(212, 160, GdkPixbuf.InterpType.BILINEAR) # @UndefinedVariable
        self["about_window"].set_logo(scaled_buf)
        self["about_window"].connect("response", on_aboutbox_response)
        self["about_window"].connect("close", on_aboutbox_close)
        self["about_window"].connect("delete_event", on_aboutbox_close)

        self["main_window"].set_icon_list(get_icon_list())

        self.reset_all_views()
        if not settings.DEBUG:
            self.get_top_widget().maximize()
        self.set_layout_mode(settings.DEFAULT_LAYOUT)

        self._clear_gdk_windows()
        self.get_top_widget().show()

        return

    def _clear_gdk_windows(self):
        gdktops = Gtk.Window.list_toplevels()
        gtktop = self["main_window"]
        our_gdktop = gtktop.get_window()        
        for gdktop in gdktops:
            if not our_gdktop == gdktop:
                gdktop.hide()

    def setup_plot(self, plot_controller):
        # Get plot canvas widget
        self.canvas_widget = plot_controller.get_canvas_widget()
        self.canvas_widget.set_name("matplotlib_box2")
        self["matplotlib_box2"] = self.canvas_widget
        
        # Get plot toolbar widget
        self.nav_toolbar = plot_controller.get_toolbar_widget(self.get_top_widget())
        self.nav_toolbar.set_name("navtoolbar")
        self["navtoolbar"] = self.nav_toolbar
        
        # Insert into the window hierarchy:
        self["matplotlib_box"].add(self.canvas_widget)
        self["matplotlib_box"].show_all()       
        
        self["navtoolbar_box"].add(self.nav_toolbar)
        self.nav_toolbar.hide()

    def reset_child_view(self, view_name, class_type=None):
        if getattr(self, view_name, None) is not None:
            getattr(self, view_name).hide()
            setattr(self, view_name, None)
        if class_type == None:
            class_type = self.child_views[view_name]
        view = class_type(parent=self)
        setattr(self, view_name, view)
        view.set_layout_mode(self.current_layout_state)

        if view_name.lower() == "project":
            # Plug in this tree view in the main application:
            self._add_child_view(
                view.specimens_treeview_container, self["specimens_container"])

        return view

    def reset_all_views(self):
        for view_name, class_type in self.child_views.items():
            self.reset_child_view(view_name, class_type)

    # ------------------------------------------------------------
    #      Sensitivity updates
    # ------------------------------------------------------------
    def update_project_sensitivities(self, project_loaded):
        """
            Updates the views sensitivities according to the flag 'project_loaded'
            indicating whether or not there's a project loaded.
        """
        self["main_pained"].set_sensitive(project_loaded)
        self["project_actions"].set_sensitive(project_loaded)
        for action in self["project_actions"].list_actions():
            action.set_sensitive(project_loaded)

    def update_specimen_sensitivities(self, single_specimen_selected, multiple_specimen_selected):
        """
            Updates the views sensitivities according to the flags 
            'single_specimen_active' indicating whether or not there's a single
            specimen selected (= active) and 'multiple_specimen_active' 
            indicating whether or not there are multiple specimen selected.
        """
        self["specimen_actions"].set_sensitive(single_specimen_selected)
        self["specimens_actions"].set_sensitive(single_specimen_selected or multiple_specimen_selected)

    def update_plot_status(self, angularpos, dspacing, experimental, calculated=None):
        wrapper = "<span font_family=\"monospace\">%s</span>"
        text = ""
        if angularpos is not None:
            text = "20=% 3.2f °    d=% 3.2f nm    I<sub>e</sub>=% 5d" % (angularpos, dspacing, experimental)
            if calculated is not None:
                text += "    I<sub>c</sub>=% 5d" % calculated
        self["lbl_plot_info"].set_markup(wrapper % text) 
        
    # ------------------------------------------------------------
    #      View update methods
    # ------------------------------------------------------------
    def set_layout_mode(self, mode):
        super(AppView, self).set_layout_mode(mode)
        for view_name in self.child_views:
            getattr(self, view_name).set_layout_mode(mode)

    def show_plot_toolbar(self):
        self.nav_toolbar.show()
        
    def hide_plot_toolbar(self):
        self.nav_toolbar.hide()

    def show(self, *args, **kwargs):
        BaseView.show(self, *args, **kwargs)

    def get_toplevel(self):
        return self["main_window"]

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on
#  -------------------------------------------------------------------------
#  Copyright (C) 2014 by Mathijs Dumon <mathijs dot dumon at gmail dot com>
#
#  mvc is a framework derived from the original pygtkmvc framework
#  hosted at: <http://sourceforge.net/projects/pygtkmvc/>
#
#  mvc is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2 of the License, or (at your option) any later version.
#
#  mvc is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110, USA.
#  -------------------------------------------------------------------------

import logging
logger = logging.getLogger(__name__)

import weakref

from .metaclasses import MetaAdapter

class AbstractAdapter(object, metaclass=MetaAdapter):
    """
        An semi-abstract class all Adapters have to derive from.
    """

    widget_types = []

    __prop = None
    @property
    def _prop(self):
        if callable(self.__prop):
            return self.__prop()
        else:
            return self.__prop
    @_prop.setter
    def _prop(self, value):
        if value is None:
            self.__prop = None
        else:
            self.__prop = weakref.ref(value, lambda: self.disconnect())

    __controller = None
    @property
    def _controller(self):
        if callable(self.__controller):
            return self.__controller()
        else:
            return self.__controller
    @_controller.setter
    def _controller(self, value):
        if value is None:
            self.__controller = None
        else:
            self.__controller = weakref.ref(value, lambda c: self.disconnect())

    __widget = None
    @property
    def _widget(self):
        if callable(self.__widget):
            return self.__widget()
        else:
            return self.__widget
    @_widget.setter
    def _widget(self, value):
        if value is None:
            self.__widget = None
        else:
            self.__widget = weakref.ref(value, lambda w: self.disconnect(widget=w()))

    # ----------------------------------------------------------------------
    #  Construction:
    # ----------------------------------------------------------------------
    def __init__(self, controller, prop, widget, *args, **kwargs):
        super(AbstractAdapter, self).__init__(*args, **kwargs)

        self._prop = prop
        self._controller = controller
        self._widget = widget

    # ----------------------------------------------------------------------
    #  Public interface:
    # ----------------------------------------------------------------------
    def update_model(self):
        """Forces the property to be updated from the value hold by
        the widget. This method should be called directly by the
        user in very unusual conditions."""
        self._write_property(self._read_widget())
        return

    def update_widget(self):
        """Forces the widget to be updated from the property
        value. This method should be called directly by the user
        when the property is not observable, or in very unusual
        conditions."""
        self._write_widget(self._read_property())
        return

    def disconnect(self, model=None, widget=None):
        """Disconnects the adapter from the model and the widget."""
        self._disconnect_model(model=model)
        self._disconnect_widget(widget=widget)

    # ----------------------------------------------------------------------
    #  Widget connecting & disconnecting:
    # ----------------------------------------------------------------------
    def _connect_widget(self):
        raise NotImplementedError("Please Implement this method")

    def _disconnect_widget(self, widget=None):
        raise NotImplementedError("Please Implement this method")

    # ----------------------------------------------------------------------
    #  Model connecting & disconnecting:
    # ----------------------------------------------------------------------
    def _connect_model(self):
        raise NotImplementedError("Please Implement this method")

    def _disconnect_model(self, model=None):
        raise NotImplementedError("Please Implement this method")

    # ----------------------------------------------------------------------
    #  Widget-side reading and writing
    # ----------------------------------------------------------------------
    def _read_widget(self):
        raise NotImplementedError("Please Implement this method")

    def _write_widget(self, val):
        raise NotImplementedError("Please Implement this method")

    # ----------------------------------------------------------------------
    #  Model-side reading and writing
    # ----------------------------------------------------------------------
    def _read_property(self, *args):
        raise NotImplementedError("Please Implement this method")

    def _write_property(self, value, *args):
        raise NotImplementedError("Please Implement this method")

    pass # end of class




import logging
logger = logging.getLogger(__name__)

import types, json
import numpy as np

from ..support.utils import pop_kwargs

from .properties.labeled_property import LabeledProperty
from .base import Model

class XYData(Model):
    """
        An XYData is a mixin model holding a list of x-y numbers.  
        Its values can be indexed, e.g.:
         >>> xydata = XYData(data=([1, 2, 3], [[4, 5], [6, 7], [8, 9]]))
         >>> xydata[0]
         (1, [4, 5])
         
        and iterated:
         >>> xydata = XYData(data=([1, 2, 3], [[4, 5], [6, 7], [8, 9]]))
         >>> for row in xydata:
         ...  print row
         ...
         (1, [4, 5])
         (2, [6, 7])
         (3, [8, 9])
         
        You can also associate names with each column:
         >>> xydata = XYData(data=([1, 2, 3], [[4, 5], [6, 7], [8, 9]]))
         >>> xydata.y_names = ["First Column", "Second Column"]
         >>> xydata.y_names.get(0, "")
         'First Column'
    """

    # OBSERVABLE PROPERTIES:

    #: The X Data
    data_x = LabeledProperty(default=None, text="X data")
    @data_x.setter
    def data_x(self, value):
        self.set_data(value, self.data_y)

    #: The Y Data
    data_y = LabeledProperty(default=None, text="Y data")
    @data_y.setter
    def data_y(self, value):
        self.set_data(self.data_x, value)

    # REGULAR PROPERTIES:
    _y_names = []
    @property
    def y_names(self):
        if len(self) < len(self._y_names):
            return self._y_names[:len(self)]
        else:
            return self._y_names
    @y_names.setter
    def y_names(self, names):
        self._y_names = names if names is not None else []

    @property
    def size(self):
        return len(self)

    @property
    def num_columns(self):
        return 1 + self.data_y.shape[1]

    @property
    def max_y(self):
        if len(self.data_y) > 1:
            return np.max(self.data_y)
        else:
            return 0

    @property
    def min_y(self):
        if len(self.data_x) > 1:
            return np.min(self.data_y)
        else:
            return 0

    @property
    def abs_max_y(self):
        if len(self.data_x) > 1:
            return np.max(np.absolute(self.data_y))
        else:
            return 0

    @property
    def abs_min_y(self):
        if len(self.data_x) > 1:
            return np.min(np.absolute(self.data_y))
        else:
            return 0

    # ------------------------------------------------------------
    #      Initialization and other internals
    # ------------------------------------------------------------
    def __init__(self, *args, **kwargs):
        """
            Valid keyword arguments for an XYData are:
                data: the actual data containing x and y values, this can be a:
                 - JSON string: "[[x1, x2, ..., xn], [y11, y21, ..., yn1], ..., [y1m, y2m, ..., ynm]]"
                 - A dictionary from a (deprecated) XYObjectListStore, containing
                   a data property, which contains a JSON string as above.
                 - A 2D-numpy array, in which its first axes contains the 
                   data rows, and its second axes contains the columns, first 
                   column being the x-data, and following columns the y-data, e.g.:
                    np.array([[x1,y11,...,y1m],
                              [x2,y21,...,y2m],
                              ...,
                              [xn,yn1,...,ynm]])
                  - An iterable containing the x-data and y-data as if it would be
                    passed to set_data(*data), e.g.:
                     ([1, 2, 3], [[4, 5], [6, 7], [8, 9]])
                names: names for the y columns (optional)
        """

        XYData.set_data(self, np.array([], dtype=float), np.zeros(shape=(0, 0), dtype=float))

        my_kwargs = pop_kwargs(kwargs,
            "names", "data",
            *[prop.label for prop in type(self).Meta.get_local_persistent_properties()]
        )
        super(XYData, self).__init__(*args, **kwargs)
        kwargs = my_kwargs

        self.y_names = kwargs.get("names", self.y_names)

        data = kwargs.get("data", None)
        if data is not None:
            if isinstance(data, str):
                XYData._set_from_serial_data(self, data)
            elif isinstance(data, dict):
                XYData._set_from_serial_data(self, data["properties"]["data"])
            elif isinstance(data, np.ndarray):
                XYData.set_data(self, data[:, 0], data[:, 1:])
            elif hasattr(data, '__iter__'):
                XYData.set_data(self, *data)

    # ------------------------------------------------------------
    #      Input/Output stuff
    # ------------------------------------------------------------

    def _serialize_data(self):
        """
            Internal method, should normally not be used!
            If you want to write data to a file, use the save_data method instead!
        """
        conc = np.insert(self.data_y, 0, self.data_x, axis=1)
        return "[" + ",".join(
                ["[" + ",".join(["%f" % val for val in row]) + "]" for row in conc]
            ) + "]"

    def _deserialize_data(self, data):
        """
            Internal method, should normally not be used!
            If you want to load data from a file, use the generic.io.file_parsers
            classes in combination with the load_data_from_generator instead!
            'data' argument should be a json string, containing a list of lists
            of x and y values, i.e.:
            [[x1, x2, ..., xn], [y11, y21, ..., yn1], ..., [y1m, y2m, ..., ynm]]
            If there are n data points and m+1 columns.
        """
        data = data.replace("nan", "'nan'")
        data = data.replace("-inf", "'-inf'")
        data = data.replace("+inf", "'+inf'")
        data = data.replace("inf", "'inf'")
        data = json.JSONDecoder().decode(data)
        return data

    def _set_from_serial_data(self, sdata):
        """Internal method, do not use!"""
        data = []
        try:
            data = self._deserialize_data(sdata)
        except ValueError:
            logger.exception("Failed to deserialize xy-data string `%s`" % sdata)
        if data != []:
            data = np.array(data, dtype=float)
            try:
                x = data[:, 0]
                y = data[:, 1:]
            except IndexError:
                logger.exception("Failed to load xy-data from serial string: %s" % sdata)
            else:
                XYData.set_data(self, x, y)

    def load_data_from_generator(self, generator, clear=True):
        if clear: self.clear()
        for x, y in generator:
            self.append(x, y)

    # ------------------------------------------------------------
    #      X-Y Data Management Methods & Functions
    # ------------------------------------------------------------
    def _y_from_user(self, y_value):
        return np.array(y_value, ndmin=2, dtype=float)

    def set_data(self, x, y):
        """
            Sets data using the supplied x, y1, ..., yn arrays.
        """
        tempx = np.asarray(x)
        tempy = np.asarray(y)
        if tempy.ndim == 1:
            tempy = tempy.reshape((tempy.size, 1))
        if tempx.shape[0] != tempy.shape[0]:
            raise ValueError("Shape mismatch: x (shape = %s) and y (shape = %s) data need to have compatible shapes!" % (tempx.shape, tempy.shape))
        self._data_x = tempx
        self._data_y = tempy

    def set_value(self, i, j, value):
        if i < len(self):
            if j == 0:
                self.data_x[i] = value
            elif j >= 1:
                self.data_y[i, j - 1] = np.array(value, dtype=float)
            else:
                raise IndexError("Column indices must be positive values (is '%d')!" % j)
        else:
            raise IndexError("Row index '%d' out of bound!" % i)

    def append(self, x, y):
        """
            Appends data using the supplied x, y1, ..., yn arrays.
        """
        data_x = np.append(self.data_x, x)
        _y = self._y_from_user(y)
        if self.data_y.size == 0:
            data_y = _y
        else:
            data_y = np.append(self.data_y, _y, axis=0)
        self.set_data(data_x, data_y)

    def insert(self, pos, x, y):
        """
            Inserts data using the supplied x, y1, ..., yn arrays at the given
            position.
        """
        self.data_x = np.insert(self.data_x, pos, x)
        self.data_y = np.insert(self.data_y, pos, self._y_from_user(y), axis=0)

    def remove_from_indeces(self, *indeces):
        if indeces != []:
            indeces = np.sort(indeces)[::-1]
            for index in indeces:
                self.set_data(
                    np.delete(self.data_x, index, axis=0),
                    np.delete(self.data_y, index, axis=0)
                )

    def clear(self):
        """
            Clears all x and y values.
        """
        self.set_data(np.zeros((0,), dtype=float), np.zeros((0, 0), dtype=float))

    # ------------------------------------------------------------
    #      Convenience Methods & Functions
    # ------------------------------------------------------------
    def get_xy_data(self, column=1):
        """
            Returns a two-tuple containing 1D-numpy arrays with the x-data and
            the y-data for a given column. If the column keyword is not passed, 
            the first column is returned.
        """
        if len(self) > 0:
            return self.data_x, self.data_y[:, column - 1]
        else:
            return np.array([], dtype=float), np.array([], dtype=float)

    def get_y_at_x(self, x, column=0):
        """ 
            Get the (interpolated) value for the y-column 'column' for
            a given x value
        """
        if self._data_x.size:
            return np.interp(x, self._data_x, self._data_y[:, column])
        else:
            return 0

    def get_y_name(self, column):
        """
            Returns the name of the given column. If the y_names attribute is 
            not properly set (e.g. too small or empty), it will return an empty
            string. This method is 'safer' to use then directly accessing the
            y_names attribute (may result in an IndexError).
        """
        try:
            return self.y_names[column]
        except IndexError:
            return ""

    # ------------------------------------------------------------
    #      Iterable & Indexable implementation
    # ------------------------------------------------------------
    def __len__(self):
        return len(self.data_x)

    def __getitem__(self, index):
        return self.data_x[index], self.data_y[index].tolist()

    def __iter__(self):
        for i in range(len(self)):
            yield self[i]

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on
#  -------------------------------------------------------------------------
#  Copyright (C) 2014 by Mathijs Dumon <mathijs dot dumon at gmail dot com>
#  Copyright (C) 2005 by Roberto Cavada <roboogle@gmail.com>
#
#  mvc is a framework derived from the original pygtkmvc framework
#  hosted at: <http://sourceforge.net/projects/pygtkmvc/>
#
#  mvc is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2 of the License, or (at your option) any later version.
#
#  mvc is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110, USA.
#  -------------------------------------------------------------------------

import types
import logging
logger = logging.getLogger(__name__)

from collections import OrderedDict

from .properties import LabeledProperty
from .object_pool import ThreadedObjectPool

from ..support.utils import get_unique_list, get_new_uuid

class ModelMeta(type):
    """
    This is a meta-class that provides support to quickly set up a model with
    observable properties. For the simplest attributes you can omit the usual
    setters and getters if nothing else needs to be done but setting and getting
    the property. You can also provide your own implementation of the setters 
    and getters (see below). Classes instantiated by this meta-class will be 
    also be registered in an object pool using a UUID attribute. You do not
    need to add this property to your inner 'Meta' class.
    Classes instantiated by this metaclass must provide a method named
    notify_property_value_change(self, prop_name, old, new), or notification of
    property changes using 'Observer' won't work.
    ModelMeta also provides multi-threading support for accessing properties, 
    through a (very basic) locking mechanism. It is assumed a lock is
    owned by the class that uses it. A Lock object called _prop_lock
    is assumed to be a member of the using class.
    
    How can you use this metaclass?
    First, '__metaclass__ = ModelMeta' must be class member of the class
    you want to make the automatic properties handling.
    
    Second, the model class needs to define an inner 'Meta' class in which
    you define meta-data which is used by this class (also see 'ModelMeta.Meta'
    class).
    This class defines several properties, the most important being 'properties'.
    This attribute is a list of descriptor objects, each of which describe
    a single property. Aside from the basics (name, default, etc.) it can
    contain additional information describing how/if other parts of the 
    framework can use this attribute (e.g. is it supposed to be stored?, does 
    it have a visual representation?). 
    That's all: after the instantiation, your class will contain all properties
    you named inside 'properties'. Each of them will be also associated
    to a couple of automatically-generated functions which get and set the
    property value inside a generated member variable.
    
    Custom setters and getters:
    Suppose the property is called 'x'.  The generated variable (which keeps
    the real value of the property x) is called _x. The getter is called 
    'get_x(self)', and the setter is called 'set_x(self, value)'. The base
    implementation of this getter is to return the value stored in the variable
    associated with the property. The setter simply sets its value.
    Programmers can override basic behavior for getters or setters simply by
    defining their getters and setters using the name convention above.
    The customized function can lie anywhere in the user classes hierarchy.
    Every overridden function will not be generated by the metaclass.

    For some properties it can be interesting to create a new descriptor class, 
    and XXXXX TODO TODO

    """

    object_pool = ThreadedObjectPool()

    # ------------------------------------------------------------
    #      Type creation:
    # ------------------------------------------------------------
    def __new__(cls, name, bases, _dict):
        # find all data descriptors, auto-set their labels
        properties = {}
        for label, _property in list(_dict.items()):
            if isinstance(_property, LabeledProperty):
                _property.label = label
                properties[label] = _property

        # Create the class type:
        new_class = super(ModelMeta, cls).__new__(cls, name, bases, _dict)
        # Get the meta class:
        meta = cls.get_meta(new_class, name, bases, _dict)
        # Set the properties for future reference:
        meta.properties = sorted(properties.values(), key=lambda prop: getattr(prop, 'declaration_index', 0))
        # Return our new class:
        return new_class

    def __init__(cls, name, bases, _dict):
        cls.process_properties(name, bases, _dict)
        return super(ModelMeta, cls).__init__(name, bases, _dict)

    # ------------------------------------------------------------
    #      Instance creation:
    # ------------------------------------------------------------
    def __call__(cls, *args, **kwargs): # @NoSelf
        """
        This method checks if the passed keyword args contained a "uuid" key, if
        so it is popped (the actual class's __init__ doesn't get it). If not
        a new UUID is created. 
        The class instance is then created and the UUID is set accordingly. 
        """
        # Check if uuid has been passed (e.g. when restored from disk)
        # if not, generate a new one
        try:
            uuid = kwargs.pop("uuid")
        except KeyError:
            uuid = get_new_uuid()

        # Create instance:
        instance = type.__call__(cls, *args, **kwargs)

        # Set the UUID
        instance.uuid = uuid

        return instance

    # ------------------------------------------------------------
    #               Services
    # ------------------------------------------------------------
    def get_meta(cls, name, bases, _dict): #@NoSelf
        """ Extracts or creates the meta class for this new model """
        try:
            meta = _dict["Meta"]
        except KeyError:
            if len(bases) == 1:
                meta = type("Meta", (bases[0].Meta,), dict(properties=[]))
                cls.set_attribute(_dict, "Meta", meta)
                _dict["Meta"] = meta
            else:
                raise TypeError("Class %s.%s has not defined an inner Meta class, and has multiple base classes!" % (cls.__module__, cls.__name__))
        return meta

    def process_properties(cls, name, bases, _dict):  # @NoSelf
        """Processes the properties defined in the class's metadata class."""

        # Get the meta class:
        meta = cls.get_meta(name, bases, _dict)

        # Get the list of properties for this class type (excluding bases):
        properties = get_unique_list(meta.properties)

        # Check the list of observables is really an iterable:
        if not isinstance(properties, list):
            raise TypeError("%s.%s.Meta 'properties' must be a list, not '%s'" %
                            (cls.__module__, cls.__name__, type(properties)))

        # Generates the list of _all_ properties available for this class's bases
        all_properties = OrderedDict()

        # Loop over bases in reverse order:
        for class_type in bases[::-1]:
            # Loop over their properties, and update the dictionary:
            if hasattr(class_type, "Meta"):
                for attr in getattr(class_type.Meta, "all_properties", []):
                    all_properties[attr.label] = attr

        # Add new/Override old attributes:
        for attr in properties:
            all_properties[attr.label] = attr

        # Set all_properties on the metadata class:
        meta.all_properties = list(all_properties.values())

        logger.debug("Class %s.%s has properties: %s" \
                     % (cls.__module__, cls.__name__, all_properties))

        pass # end of method

    def check_value_change(cls, old, new): # @NoSelf
        """Checks whether the value of the property changed in type
        or if the instance has been changed to a different instance.
        If true, a call to model._reset_property_notification should
        be called in order to re-register the new property instance
        or type"""
        from ..support import observables
        return  type(old) != type(new) or \
               isinstance(old, observables.ObsWrapperBase) and (old != new)

    def set_attribute(cls, _dict, name, value): # @NoSelf
        """Sets an attribute on the class and the dict"""
        _dict[name] = value
        setattr(cls, name, value)

    def del_attribute(cls, _dict, name): # @NoSelf
        """Deletes an attribute from the class and the dict"""
        del _dict[name]
        delattr(cls, name)

    pass # end of class
# ----------------------------------------------------------------------



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from mvc.models.metaclasses import ModelMeta
from .models import RefinementInfo

class PyXRDRefinableMeta(ModelMeta):
    """
        A metaclass for regular mvc Models with refinable properties.
    """

    # ------------------------------------------------------------
    #      Instance creation:
    # ------------------------------------------------------------
    def __call__(cls, *args, **kwargs):   # @NoSelf
        """
        Strips the refinement info data from the keyword argument dictionary,
        passes the stripped dictionary to create the actual class instance,
        creates the attributes on the instance and returns it.
        """
        # Pop & parse any refinement info keyword arguments that might be present:
        prop_infos = dict()
        for prop in cls.Meta.all_properties:
            if getattr(prop, "refinable", False):
                ref_info_name = prop.get_refinement_info_name()
                info_args = kwargs.pop(ref_info_name, None)
                if info_args:
                    prop_infos[ref_info_name] = RefinementInfo.from_json(*info_args)
                else:
                    prop_infos[ref_info_name] = RefinementInfo(prop.minimum, prop.maximum, False)

        # Create the instance passing the stripped keyword arguments:
        instance = ModelMeta.__call__(cls, *args, **kwargs)

        # Set the refinement attributes on the newly created instance:
        for ref_info_name, ref_info in prop_infos.items():
            setattr(instance, ref_info_name, ref_info)

        # Return the instance:
        return instance

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on
#  -------------------------------------------------------------------------
#  Copyright (C) 2014 by Mathijs Dumon <mathijs dot dumon at gmail dot com>
#
#  mvc is a framework derived from the original pygtkmvc framework
#  hosted at: <http://sourceforge.net/projects/pygtkmvc/>
#
#  mvc is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2 of the License, or (at your option) any later version.
#
#  mvc is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110, USA.
#  -------------------------------------------------------------------------

from mvc.support.utils import  rec_getattr, rec_setattr
from mvc.support.observables.value_wrapper import ValueWrapper
import inspect

class Mixable(object):
    """
    Base class that allows to mix-in any other new-style class
    passed in the constructor, using the mix_with keyword argument
    (tuple of class types).
    """

    def __init__(self, mix_with=None, *args, **kwargs):
        super(Mixable, self).__init__(*args, **kwargs)
        # Nifty hook so we can mix-in any other class:
        if mix_with is not None:
            name = type(self).__name__
            for klass in mix_with:
                name = "%s_%s" % (klass.__name__.replace("Mixin", ""), name)
            bases = tuple(mix_with) + (type(self),)
            self.__class__ = type(name, bases, {})

    pass #end of class

class LabeledProperty(Mixable):
    """
     Property descriptor base class to be used in combination with a
     ~:class:`mvc.models.Model` (sub)class.
     Expects it's label (attribute name) to be set or passed to __init__, for
     ~:class:`mvc.models.Model` (sub)class this is done automagically using
     its metaclass.
     
     Additional keyword arguments will be set as attributes on the descriptor.
     
     Some of these keywords have been given sane default values, even though
     they are not required for the implementation:
     
         - title = label
         - math_title = label
         - description = label
         - persistent (False)
         - store_private = (False)
         - visible (False)
         - data_type (object)
         - observable (True)
         - widget_type ('custom')
         
     To use this class, use it like the regular Property or property decorators.
     E.g.:
         attribute = LabeledProperty(...)
     or
         @LabeledProperty(...)
         def get_attribute(self):
             return self.attribute
             
     The setter function is expected to accept either a single argument (the 
     new value) or two arguments (the property descriptor instance and the new 
     value). Similarly, the getter and deleter function are expected to accept
     no arguments or a single argument (the property descriptor function instance).
    """

    #: The actual attribute name used to access the value on the class instance.
    _label = None
    @property
    def label(self): return self._label
    @label.setter
    def label(self, value):
        self._label = value
        # Wrap the underlying variable if needed
        # (e.g. if it's a list, tuple, dict, or other mutable class):
        if self.default is not None:
            self.default = ValueWrapper.wrap_value(self.label, self.default, verbose=bool(value == "specimens"))

    #: Whether this attribute is persistent (needs to be stored)
    persistent = False

    #: Either False (use the actual attribute), True (use the private attribute value) or
    #: the name of the attribute to use when storing this attribute (if persistent = True)
    store_private = False

    #: A short textual description of this attribute
    title = None

    #: A short MathText description of this attribute
    math_title = None

    #: A flag indicating whether this attribute is visible in a GUI
    visible = False

    #: A string describing what kind of GUI widget can be used to display the
    #: contents of this property. Check the :mod:`mvc.adapters` module.
    widget_type = 'custom'

    #: A flag indicating whether this attribute should be visible in a table GUI
    tabular = False

    #: The type of the value stored in the attribute, or in the case of a
    #: collection (i.e. lists, dicts and tuples) the type of the collection
    #: items.
    data_type = object

    #: A flag indicating whether this attribute can be observed for changes
    observable = True

    #: The default value for this property
    default = None

    #: Declaration index: used to allow sorting in order of declaration
    declaration_index = 0

    ############################################################################
    #    Generic private attribute getters and setters
    ############################################################################
    private_attribute_format = "_%(label)s"
    def _get_private_label(self):
        """ Private attribute label (holds the actual value on the model) """
        return self.private_attribute_format % { 'label': self.label }

    def _set(self, instance, value):
        """ Private setter """
        rec_setattr(instance, self._get_private_label(), value)

    def _get(self, instance):
        """ Private getter """
        return rec_getattr(instance, self._get_private_label(), self.default)

    ############################################################################
    #    Decorator calls:
    ############################################################################
    def __call__(self, fget):
        return self.getter(fget)

    def _inject_self(self, f):
        """ Injects self into the arguments of function `f`
           (first argument after self) """
        def wrapper(*args, **kwargs):
            return f(*(args[0] + (self,) + args[1:]), **kwargs)
        return wrapper

    def getter(self, fget):
        # Getter expects to be passed the descriptor
        if len(inspect.getargspec(fget).args) > 1:
            fget = self._inject_self(fget)
        self.fget = fget
        return self

    def setter(self, fset):
        # Setter expects to be passed the descriptor
        if len(inspect.getargspec(fset).args) > 2:
            fset = self._inject_self(fset)
        self.fset = fset
        return self

    def deleter(self, fdel):
        # Deleter expects to be passed the descriptor
        if len(inspect.getargspec(fdel).args) > 1:
            fdel = self._inject_self(fdel)
        self.fdel = fdel
        return self

    ############################################################################
    #    Initialization:
    ############################################################################
    def __init__(self, fget=None, fset=None, fdel=None, doc=None, default=None, label=None, mix_with=None, **kwargs):
        super(LabeledProperty, self).__init__(mix_with=mix_with)

        #Increment declaration counter
        LabeledProperty.declaration_index += 1
        self.declaration_index = LabeledProperty.declaration_index

        self.fget = fget
        self.fset = fset
        self.fdel = fdel
        if doc is None and fget is not None:
            doc = fget.__doc__
        self.__doc__ = doc

        self.label = self.math_title = self.title = self.description = label
        self.persistent_label = label
        self.default = default

        for key, value in kwargs.items():
            setattr(self, key, value)

    ############################################################################
    #    Comparison protocol:
    ############################################################################
    def __eq__(self, other):
        # Descriptors are equal if they describe the same attribute
        return other is not None and self.label == other.label

    def __hash__(self):
        return hash(self.label)

    def __neq__(self, other):
        # Descriptors are equal if they describe the same attribute
        return not self.__eq__(other)

    ############################################################################
    #    Descriptor protocol:
    ############################################################################
    def __get__(self, instance, owner):
        if instance is None:
            return self

        with instance._prop_lock:
            if self.fget is None:
                return self._get(instance)
            else:
                return self.fget(instance)

    def __set__(self, instance, value):
        with instance._prop_lock:

            # Get the old value
            old = getattr(instance, self.label)

            # Wrap the new value
            value = ValueWrapper.wrap_value(self.label, value, instance)

            # Set the new value
            if self.fset is None:
                self._set(instance, value)
            else:
                self.fset(instance, value)

            # Notify any observers
            if self.observable:
                # Check if we've really changed it, and send notifications if so:
                if type(instance).check_value_change(old, value):
                    instance._reset_property_notification(self, old)
                    pass

                # Notify any interested party we have set this property!
                if hasattr(instance, 'notify_property_value_change'):
                    instance.notify_property_value_change(self.label, old, value)

    def __delete__(self, instance):
        if self.fdel is None:
            raise AttributeError("Can't delete attribute `%s`!" % self.label)
        self.fdel(instance)

    def __repr__(self):
        return '<%s %s>' % (self.__class__.__name__, self.label)

    pass #end of class
